# Copyright (c) 2020 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import random
import re

from oslo_log import log

import tobiko
from tobiko.openstack import topology
from tobiko.shell import sh
from tobiko.tripleo import pacemaker


LOG = log.getLogger(__name__)


def get_filtered_node_containers(node, containers_regex):
    """Search all containers are matched with containers_regex list

    :param node: Node to search containers on
    :type node: class: tobiko.openstack.topology.OpenStackTopologyNode
    :param containers_regex: List of regex that can be matched to containers
    :type containers_regex: list of strings
    :return: List of contatiner names
    :rtype: list of strings
    """
    filtered_containers = []
    # 'docker' is used here in order to be compatible with old OSP versions.
    # On versions with podman, 'docker' command is linked to 'podman'
    result = sh.execute(
            'docker ps --format "{{.Names}}"',
            ssh_client=node.ssh_client, sudo=True)
    all_node_containers = result.stdout.strip().split('\n')
    for container in all_node_containers:
        container = container.strip('"')
        if any(re.fullmatch(reg, container) for reg in containers_regex):
            filtered_containers.append(container)
    return filtered_containers


def get_nodes_for_groups(groups):
    """Search for all nodes that are matched with the specified groups

    :param groups: List of groups nodes can belong to
    :type groups: list
    :return: List of nodes that belong to the specified groups
    :rtype: list of tobiko.openstack.topology.OpenStackTopologyNode
    """
    nodes = []
    for node in topology.list_openstack_nodes():
        for group in node.groups:
            if group in groups:
                nodes.append(node)
    return(nodes)


def get_config_files(node, kolla_jsons, conf_ignorelist, scripts_to_check):
    """Return the list of configuration files according to kolla JSONs

    Kolla has the container execution commands in the matched JSON files.
    Need to get all the '--config-file <file_name>' parameters from those
    JSON files.
    It is possible that the container is executed with the set of commands
    are written in a separate script instead of a single command so need
    to follow those scripts to get configuration files from there too.

    :param node: Node to search configuration files on
    :type node: class: tobiko.openstack.topology.OpenStackTopologyNode
    :param kolla_jsons: Path to kolla json files to get commands from
    :type kolla_jsons: list
    :param conf_ignorelist: Configuration files to ignore
    :type conf_ignorelist: list
    :param scripts_to_check: Sripts to look for configuration files in if those
        scripts will be found in kolla json files.
    :type scripts_to_check: list
    :return: List of config files paths within containers
    :rtype: list
    """
    cmds = sh.execute(
            f"jq '.command' {' '.join(kolla_jsons)}",
            ssh_client=node.ssh_client,
            expect_exit_status=None, sudo=True).stdout.strip().split('\n')
    LOG.debug(f'{node.name} run containers with commands {cmds}')
    config_files = set()
    for cmd in cmds:
        cmd = cmd.strip('"')
        if cmd in scripts_to_check:
            LOG.debug(f'{cmd} is recognized as script to search '
                      'for config files in')
            oc_script_location = sh.execute(
                    f'find /var/lib | grep {cmd} | grep -v overlay',
                    ssh_client=node.ssh_client,
                    sudo=True).stdout.strip().split('\n')[0]
            cmd = sh.execute(
                    f'cat {oc_script_location}',
                    ssh_client=node.ssh_client, sudo=True).stdout.strip()
            cmd = cmd.strip('"')
        temp_conf_files = re.findall('--config-file [^ \n]*', cmd)
        for conf_file in temp_conf_files:
            conf_file = conf_file.split(' ')[1]
            if conf_file in conf_ignorelist:
                LOG.debug(f'{conf_file} is in ignore list')
                continue
            config_files.add(conf_file)
    LOG.debug(f'There are {config_files} on {node.name}')
    return config_files


def get_node_neutron_containers(node):
    """Return list of all neutron containers are available on the node

    :param node: Node to search containers on
    :type node: class: tobiko.openstack.topology.OpenStackTopologyNode
    :return: List of neutron containers names
    :rtype: list of strings
    """
    neutron_containers = ['neutron_((ovs|metadata|l3)_agent|dhcp|api)',
                          'ovn_metadata_agent']
    return get_filtered_node_containers(node, neutron_containers)


def get_node_ovn_containers(node):
    """Return list of all ovn containers are available on the node

    :param node: Node to search containers on
    :type node: class: tobiko.openstack.topology.OpenStackTopologyNode
    :return: List of neutron containers names
    :rtype: list of strings
    """
    neutron_containers = [
            'ovn_(controller|metadata_agent)',
            r'ovn-dbs-bundle-(podman|docker)-\d*']
    return get_filtered_node_containers(node, neutron_containers)


def get_node_neutron_config_files(node):
    """Return all relevant neutron config files

    :param node: Overcloud node to search log files on
    :type node: class: tobiko.openstack.topology.OpenStackTopologyNode
    :return: List of config files paths within neutron containers
    :rtype: list of strings
    """
    kolla_jsons = ['/var/lib/kolla/config_files/neutron*']
    conf_ignorelist = ['/usr/share/neutron/neutron-dist.conf']
    scripts_to_check = ['/neutron_ovs_agent_launcher.sh']
    config_files = get_config_files(node,
                                    kolla_jsons,
                                    conf_ignorelist,
                                    scripts_to_check)
    return config_files


def get_node_ovn_config_files():
    """Return all relevant ovn config files

    :return: List of config files paths within ovn containers
    :rtype: list of strings
    """
    return ['/etc/openvswitch/default.conf']


def get_pacemaker_resource_from_container(container):
    """Returns pacemaker resource name or None

    :param container: Name of the container
    :type container: string
    :return: pacemaker resource name or None
    :rtype: string
    """
    pcs_resource = None
    resource_candidate = re.search(r'(.*)-(docker|podman)-\d', container)
    if resource_candidate:
        pcs_resource = resource_candidate.group(1)
    return pcs_resource


def get_node_logdir_from_pcs(node, container):
    """ Return the logdir for a given pacemaker resource

    :param node: Node the container is running on
    :type node: class: tobiko.openstack.topology.OpenStackTopologyNode
    :param container: Name of the container
    :type container: string
    :return: Path to the logfiles on the container
    :rtype: string
    """
    pcs_resource = get_pacemaker_resource_from_container(container)
    if pcs_resource is None:
        return
    logdir = None
    out_lines = pacemaker.run_pcs_resource_operation(
            pcs_resource, pacemaker.SHOW, node=node.ssh_client).splitlines()
    log_files_regex = re.compile(
        r'^\s*options=.*source-dir=(.*) target-dir=.*-log-files\)$')
    for line in out_lines:
        log_files_match = log_files_regex.search(line)
        if log_files_match:
            logdir = log_files_match.group(1)
            if line.endswith('-new-log-files)'):
                break
            else:
                continue
    return logdir


def get_pacemaker_resource_logfiles(node, container):
    logfiles = []
    exclude_pid_files = 'ovn-controller.pid'
    resource = get_pacemaker_resource_from_container(container)
    out_lines = pacemaker.run_pcs_resource_operation(
        resource, pacemaker.SHOW, node=node.ssh_client).splitlines()
    run_files_regex = re.compile(
        r'^\s*options=.*source-dir=(.*) target-dir=.*-run-files\)$')
    for line in out_lines:
        run_files_match = run_files_regex.search(line)
        if run_files_match:
            pid_files = (sh.execute(f'find {run_files_match.group(1)} '
                                    f'-name *.pid ! -name {exclude_pid_files}',
                                    ssh_client=node.ssh_client).
                         stdout.splitlines())
            break
    pids = sh.execute(f'cat {" ".join(pid_files)}',
                      ssh_client=node.ssh_client,
                      sudo=True).stdout.splitlines()
    for pid in pids:
        cmd_stdout = sh.execute(f'docker exec -u root {container} '
                                f'cat /proc/{pid}/cmdline',
                                ssh_client=node.ssh_client,
                                sudo=True).stdout
        for log_file in re.findall('--log-file=[^ \n\x00]*', cmd_stdout):
            logfiles.append(log_file.split('=')[1])
    return logfiles


def get_default_container_logfiles(container):
    CONTAINER_LOGFILE_DICT_LIST = [
        {'container_regex': 'ovn_controller',
         'default_logfiles': ['/var/log/openvswitch/ovn-controller.log']}]
    for cont_logfile_dict in CONTAINER_LOGFILE_DICT_LIST:
        if re.fullmatch(cont_logfile_dict['container_regex'], container):
            return cont_logfile_dict['default_logfiles']
    raise RuntimeError(f'No default log file found for container {container}')


def get_container_logfiles(node, container):
    """ Return the logfiles of the processes that are executed on the container

    :param node: Node the container is running on
    :type node: class: tobiko.openstack.topology.OpenStackTopologyNode
    :param container: Name of the container
    :type container: string
    :return: Path to the logfiles on the container
    :rtype: list
    """
    cmd = sh.execute(
            f'docker exec -u root {container} cat /run_command',
            ssh_client=node.ssh_client, sudo=True)
    cmd_stdout = cmd.stdout.strip()
    if 'pacemaker_remoted' in cmd_stdout:
        return get_pacemaker_resource_logfiles(node, container)
    if ' ' not in cmd_stdout:  # probably script as no space in the command
        cmd = sh.execute(
                f'docker exec -u root {container} cat {cmd_stdout}',
                ssh_client=node.ssh_client, sudo=True)
        cmd_stdout = cmd.stdout.strip()
    LOG.debug(f'The following command is executed in {container} container '
              f'on {node.name} node:\n{cmd_stdout}')

    log_files = []
    for log_file in re.findall('--log-file=[^ \n]*', cmd_stdout):
        log_files.append(log_file.split('=')[1])
    if log_files:
        return log_files

    if re.findall(r'--log-file[^=]*$', cmd_stdout):
        LOG.debug(f'Using default log file for the command: {cmd_stdout}')
        return get_default_container_logfiles(container)
    LOG.warning(f'No log found for the command: {cmd_stdout}')
    return None


def log_random_msg(node, container, logfile):
    """Print random message to the container log file

    :param node: Node the container is running on
    :type node: class: tobiko.openstack.topology.OpenStackTopologyNode
    :param container: Name of the container
    :type container: string
    :param logfile: Path to the logfile on the container
    :type logfile: string
    :return: Message that has been printed to container log file
    :rtype: string
    """
    symbols = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
    random_msg = ''.join(random.choice(symbols) for i in range(30))
    LOG.debug(f'Trying to print {random_msg} string to {logfile} log file '
              f'in {container} container on {node.name} node')
    log_msg(node, container, logfile, random_msg)
    return random_msg


def log_msg(node, container, logfile, msg):
    """Print random message to the container log file

    :param node: Node the container is running on
    :type node: class: tobiko.openstack.topology.OpenStackTopologyNode
    :param container: Name of the container
    :type container: string
    :param logfile: Path to the logfile on the container
    :type logfile: string
    :param msg: Message to log
    :type msg: string
    """
    cmd = f"sh -c 'echo {msg} >> {logfile}'"
    sh.execute(f'docker exec -u root {container} {cmd}',
               ssh_client=node.ssh_client, sudo=True)


def find_msg_in_file(node, logfile, message, rotated=False):
    """Search for the message in the logfile

    :param node: Node the container is running on
    :type node: class: tobiko.openstack.topology.OpenStackTopologyNode
    :param logfile: Path of the logfile
    :type logfile: string
    :param message: Message to search for
    :type message: string
    :param rotated: Variable to flag that log file has to be rotated
        so the name will be ended by '.1'
    :type rotated: bool
    :return: True if message exists in file or False otherwise
    :rtype: bool
    """
    if rotated:
        suffix = ".1"
    else:
        suffix = ""
    LOG.debug(f'Searching for {message} in {logfile}{suffix} on {node.name}')
    result = sh.execute(f'grep -h {message} {logfile}{suffix}',
                        ssh_client=node.ssh_client,
                        expect_exit_status=None, sudo=True)
    if result.stderr:
        tobiko.fail(f'Failed to read {logfile} on {node.name}:\n'
                    f'{result.stderr}')
    elif result.stdout.strip() == message:
        return True
    else:
        return False


def rotate_logs(node):
    """Rotate all the container logs using 'logrotate'

    :param node: Node to rotate logs on
    :type node: class: tobiko.openstack.topology.OpenStackTopologyNode
    """
    containers = get_filtered_node_containers(node, ['logrotate.*', ])
    if not containers:
        tobiko.skip_test('No logrotate container has been found')
    container = containers[0]
    sh.execute(f'docker exec -u root {container} logrotate '
               '-f /etc/logrotate-crond.conf',
               ssh_client=node.ssh_client, sudo=True)


def has_docker():
    return get_docker_version() is not None


skip_unless_has_docker = tobiko.skip_unless(
    "requires docker on controller nodes", has_docker)


def get_docker_version():
    # use a fixture to save the result
    return tobiko.setup_fixture(
        DockerVersionFixture).docker_version


class DockerVersionFixture(tobiko.SharedFixture):

    docker_version = None

    def setup_fixture(self):
        controller = topology.find_openstack_node(group='controller')
        try:
            result = sh.execute('docker --version',
                                ssh_client=controller.ssh_client)
        except sh.ShellCommandFailed:
            pass
        else:
            self.docker_version = result.stdout



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing
from collections import abc
import random
import re

import testtools

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import neutron
from tobiko.openstack import nova
from tobiko.openstack import topology as osp_topology
from tobiko.shell import ping
from tobiko.shell import sh


PatternType = type(re.compile(r''))


@keystone.skip_unless_has_keystone_credentials()
class OpenStackTopologyTest(testtools.TestCase):

    expected_group: osp_topology.OpenstackGroupNamesType = None

    @property
    def topology(self) -> osp_topology.OpenStackTopology:
        return osp_topology.get_openstack_topology()

    def test_get_openstack_topology(self):
        topology_class = type(self.topology)
        topo = osp_topology.get_openstack_topology(
            topology_class=topology_class)
        self.assertIs(topo, self.topology)
        self.assertIsInstance(topo, osp_topology.OpenStackTopology)

    def test_ping_node(self):
        for node in self.topology.nodes:
            ping.ping(node.public_ip, count=1, timeout=5.).assert_replied()

    def test_ssh_client(self):
        for node in self.topology.nodes:
            self.assertIsNotNone(node.ssh_client)
            hostname = sh.ssh_hostname(
                ssh_client=node.ssh_client).split('.')[0]
            self.assertEqual(node.name, hostname)

    def test_controller_group(self):
        nodes = list(self.topology.get_group('controller'))
        self.assertNotEqual([], nodes)
        for node in nodes:
            self.assertIn('controller', node.groups)

    def test_compute_group(self):
        nodes = list(self.topology.get_group('compute'))
        self.assertNotEqual([], nodes)
        for node in nodes:
            self.assertIn('compute', node.groups)
        hypervisors = {
            hypervisor.hypervisor_hostname.split('.', 1)[0].lower(): hypervisor
            for hypervisor in nova.list_hypervisors()}
        for name, hypervisor in hypervisors.items():
            node = self.topology.get_node(name)
            self.assertEqual(name, node.name)
            self.assertIn(node, nodes)

    def test_list_openstack_topology(self, group=None, hostnames=None):
        nodes = osp_topology.list_openstack_nodes(
            topology=self.topology, group=group, hostnames=hostnames)
        self.assertTrue(set(nodes).issubset(set(self.topology.nodes)))
        self.assertEqual(len(set(nodes)), len(nodes),
                         f"Repeated node found: {nodes}")
        for node in nodes:
            if isinstance(group, str):
                self.assertIn(group, node.groups)
            elif isinstance(group, PatternType):
                for actual_group in node.groups:
                    if group.match(actual_group):
                        break
                else:
                    self.fail(f"Any node {node.name} group matches "
                              f"'{group}': {node.groups}")
            elif isinstance(group, abc.Iterable):
                matching_groups = set(group) & set(node.groups)
                self.assertNotEqual(set(), matching_groups,
                                    f"Any group of node {node.name} "
                                    f"matches '{group}': {node.groups}")
            if hostnames:
                hostnames = [node_name_from_hostname(h)
                             for h in hostnames]
                self.assertIn(node.name, hostnames)
        return nodes

    def test_list_openstack_topology_with_group(self):
        group = self.topology.groups[0]
        expected_nodes = set(self.topology.get_group(group))
        actual_nodes = set(self.test_list_openstack_topology(group=group))
        self.assertEqual(expected_nodes, actual_nodes)

    def test_list_openstack_topology_with_group_pattern(self):
        groups = list(self.topology.groups)[:2]
        pattern = re.compile('|'.join(groups))
        expected_nodes = set()
        for group in groups:
            expected_nodes.update(self.topology.get_group(group))
        actual_nodes = set(self.test_list_openstack_topology(group=pattern))
        self.assertEqual(expected_nodes, actual_nodes)

    def test_list_openstack_topology_with_groups(self):
        groups = list(self.topology.groups)[:2]
        expected_nodes = set()
        for group in groups:
            expected_nodes.update(self.topology.get_group(group))
        actual_nodes = set(self.test_list_openstack_topology(group=groups))
        self.assertEqual(expected_nodes, actual_nodes)

    def test_list_openstack_topology_with_hostnames(self):
        expected_nodes = self.topology.nodes[0::2]
        actual_nodes = self.test_list_openstack_topology(
            hostnames=[node.name for node in expected_nodes])
        self.assertEqual(expected_nodes, actual_nodes)

    def test_list_nodes_processes(self):
        node = random.choice(osp_topology.list_openstack_nodes(
            group=self.expected_group))
        filename = sh.execute(
            'mktemp', ssh_client=node.ssh_client).stdout.strip()
        self.addCleanup(sh.execute, f"rm -f '{filename}'",
                        ssh_client=node.ssh_client)
        command_line = sh.shell_command(f"tail -F '{filename}'")
        process = sh.process(command_line,
                             ssh_client=node.ssh_client)

        # Process isn't listed before creation
        processes = osp_topology.list_nodes_processes(
            command_line=command_line,
            hostnames=[node.hostname])
        self.assertEqual([], processes,
                         'Process listed before executing it')

        # Process is listed after creation
        process.execute()
        self.addCleanup(process.kill)
        processes = osp_topology.list_nodes_processes(
            command_line=command_line,
            hostnames=[node.hostname])
        self.assertEqual(command_line, processes.unique.command_line)
        self.assertIs(node.ssh_client, processes.unique.ssh_client)

        # Process isn't listed after kill
        processes.unique.kill()
        for attempt in tobiko.retry(timeout=30., interval=5.):
            processes = osp_topology.list_nodes_processes(
                command_line=command_line,
                hostnames=[node.hostname])
            if not processes:
                break
            if attempt.is_last:
                self.fail("Process listed after killing it")

    @neutron.skip_unless_is_ovs()
    def test_l3_agent_mode(self):
        for node in osp_topology.list_openstack_nodes(
                group=['controller', 'compute', 'overcloud']):
            assert isinstance(node, osp_topology.OpenStackTopologyNode)
            self.assertEqual(
                neutron.get_l3_agent_mode(
                    l3_agent_conf_path=node.l3_agent_conf_path,
                    connection=node.connection),
                node.l3_agent_mode)
            self.assertIs(node.l3_agent_mode, node.l3_agent_mode)


def node_name_from_hostname(hostname):
    return hostname.split('.', 1)[0].lower()


@keystone.skip_unless_has_keystone_credentials()
class HostsNamespaceTest(testtools.TestCase):

    @property
    def all_hostnames(self) -> typing.List[str]:
        nodes = osp_topology.list_openstack_nodes()
        return sorted(node.hostname for node in nodes)

    def selected_hostames(self, hostnames: typing.Iterable[str] = None) -> \
            typing.Set[str]:
        if hostnames is None:
            return set(self.all_hostnames)
        else:
            return set(hostnames)

    def test_get_hosts_namespaces(self,
                                  hostnames: typing.Iterable[str] = None):
        namespaces = osp_topology.get_hosts_namespaces(hostnames=hostnames)
        self.assertIsInstance(namespaces, dict)
        for namespace, _hostnames in namespaces.items():
            self.assertIsInstance(namespace, str)
            self.assertIsInstance(_hostnames, list)
            self.assertEqual(
                set(_hostnames),
                self.selected_hostames(hostnames) & set(_hostnames))

    def test_get_hosts_namespaces_with_hostnames(self):
        self.test_get_hosts_namespaces(hostnames=self.all_hostnames[:1])

    def test_assert_namespace_in_hosts(self,
                                       hostnames: typing.Iterable[str] = None):
        namespaces = osp_topology.get_hosts_namespaces(hostnames=hostnames)
        for namespace, hostnames in namespaces.items():
            osp_topology.assert_namespace_in_hosts(namespace,
                                                   hostnames=hostnames)

    def test_assert_namespace_in_hosts_with_hostnames(self):
        self.test_assert_namespace_in_hosts(hostnames=self.all_hostnames[:1])

    def test_assert_namespaces_in_host_failure(self):
        self.assertRaises(self.failureException,
                          osp_topology.assert_namespace_in_hosts,
                          '<invalid>')

    def test_assert_namespace_not_in_hosts(
            self, hostnames: typing.Iterable[str] = None):
        osp_topology.assert_namespace_not_in_hosts('<invalid>',
                                                   hostnames=hostnames)

    def test_assert_namespace_not_in_hosts_with_hostnames(self):
        self.test_assert_namespace_not_in_hosts(
            hostnames=self.all_hostnames[:1])

    def test_assert_namespace_not_in_hosts_failure(self):
        namespaces = osp_topology.get_hosts_namespaces()
        for namespace in namespaces:
            self.assertRaises(self.failureException,
                              osp_topology.assert_namespace_not_in_hosts,
                              namespace)



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import re
import time

import openshift_client as oc
from oslo_log import log

import tobiko
from tobiko import config
from tobiko import podified
# from tobiko.openstack import glance
# from tobiko.openstack import keystone
# from tobiko.openstack import neutron
# from tobiko.openstack import stacks
# from tobiko.openstack import tests
# from tobiko.openstack import topology
# from tobiko.tests.faults.ha import test_cloud_recovery
# from tobiko.shell import ping
# from tobiko.shell import sh

CONF = config.CONF
LOG = log.getLogger(__name__)


@podified.skip_if_not_podified
def kill_all_galera_services():
    """kill all galera processes,
    check in pacemaker it is down"""
    galera_pods_num = sum(
        1 for node_name in oc.selector('nodes').qnames()
        for pod_obj in oc.get_pods_by_node(node_name)
        if 'cell1-galera' in pod_obj.fqname()
    )
    for i in range(galera_pods_num):
        oc.selector('pod/openstack-cell1-galera-{}'.format(i)).object()\
            .execute(['sh', '-c', 'kill -9 $(pidof mysqld)'],
                     container_name='galera')
        LOG.info('kill galera cell-{}'.format(i))

    retry = tobiko.retry(timeout=30, interval=5)
    for _ in retry:
        try:
            # checks wsrep cluster size is now unavailable
            result = oc.selector('pod/openstack-cell1-galera-0').object(
            ).execute(['sh', '-c', """mysql -u root --password=12345678
            -e 'SHOW STATUS LIKE "wsrep_cluster_size"'"""])
            # Capture and filter the error output
            error_output = result.err()
            non_error_message = """
            Defaulted container "galera" out of: galera,
            mysql-bootstrap (init)\n"""
            filtered_err_output = error_output.replace(non_error_message, '')
            if not filtered_err_output.strip():
                continue
        except oc.OpenShiftPythonException:
            LOG.info('all galera cells down')
            break
    time.sleep(60)
    for _ in retry:
        try:
            if int(re.search(r'wsrep_cluster_size\s+(\d+)', oc.selector(
                    'pod/openstack-cell1-galera-0').object().execute(
                ['sh', '-c', """mysql -u root --password=12345678 -e 'SHOW
                STATUS LIKE "wsrep_cluster_size"'"""], container_name='galera'
            ).out()).group(1)) == galera_pods_num:
                LOG.info('all galera cells are restored')
                return
        except oc.OpenShiftPythonException:
            continue
        return False


@podified.skip_if_not_podified
def remove_all_grastate_galera():
    """shut down galera properly,
    remove all grastate"""
    galera_pods_num = sum(
        1 for node_name in oc.selector('nodes').qnames()
        for pod_obj in oc.get_pods_by_node(node_name)
        if 'cell1-galera' in pod_obj.fqname()
    )
    for i in range(galera_pods_num):
        oc.selector('pod/openstack-cell1-galera-{}'.format(i)).object()\
            .execute(['sh', '-c', 'rm -rf /var/lib/mysql/grastate.dat '],
                     container_name='galera')
        LOG.info('delete grastate.dat cell-{}'.format(i))
    for i in range(galera_pods_num):
        oc.selector('pod/openstack-cell1-galera-{}'.format(i)).object()\
            .execute(['sh', '-c', 'kill -9 $(pidof mysqld)'],
                     container_name='galera')
        LOG.info('kill galera cell-{}'.format(i))
    retry = tobiko.retry(timeout=30, interval=5)
    for _ in retry:
        try:
            # checks wsrep cluster size is now unavailable
            result = oc.selector('pod/openstack-cell1-galera-0').object(
            ).execute(['sh', '-c', """mysql -u root --password=12345678
            -e 'SHOW STATUS LIKE "wsrep_cluster_size"'"""])
            # Capture and filter the error output
            error_output = result.err()
            non_error_message = """
            Defaulted container "galera" out of: galera,
            mysql-bootstrap (init)\n"""
            filtered_err_output = error_output.replace(non_error_message, '')
            if not filtered_err_output.strip():
                continue
        except oc.OpenShiftPythonException:
            LOG.info('all galera cells down')
            break
    time.sleep(60)
    for _ in retry:
        try:
            if int(re.search(r'wsrep_cluster_size\s+(\d+)', oc.selector(
                    'pod/openstack-cell1-galera-0').object().execute(
                ['sh', '-c', """mysql -u root --password=12345678 -e 'SHOW
                STATUS LIKE "wsrep_cluster_size"'"""], container_name='galera'
            ).out()).group(1)) == galera_pods_num:
                LOG.info('all galera cells are restored')
                return
        except oc.OpenShiftPythonException:
            continue
        return False



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

from oslo_log import log

import tobiko
from tobiko import tripleo
from tobiko import config
from tobiko.openstack.octavia import _client
from tobiko.openstack import nova, openstacksdkclient
from tobiko.openstack.octavia import _validators
from tobiko.openstack import topology
from tobiko.shell import sh

LOG = log.getLogger(__name__)
CONF = config.CONF

AmphoraType = typing.Dict[str, typing.Any]
AmphoraIdType = typing.Union[str, AmphoraType]


def get_os_conn():
    """Get openstacksdk client fixture
    """
    return openstacksdkclient.openstacksdk_client()


def get_amphora(amp_id: str):
    return get_os_conn().load_balancer.get_amphora(amp_id)


def list_amphorae(load_balancer_id: str = None, **params):
    if load_balancer_id is not None:
        params['load_balancer_id'] = load_balancer_id
    return get_os_conn().load_balancer.amphorae(
        loadbalancer_id=load_balancer_id)


def get_amphora_compute_node(load_balancer_id: str,
                             port: int,
                             protocol: str,
                             ip_address: str,
                             client: _client.OctaviaClientType = None) -> (
        topology.OpenStackTopologyNode):
    """Gets the compute node which hosts the LB amphora

    This function finds the Overcloud compute node which
    hosts the amphora. In case there are more than 1 amphora
    (e.g. if the LB's topology is Active/standby), so the compute node which
    hosts the master amphora will be returned.

    :param load_balancer_id: the load balancer ID.
    :param port: the load balancer port.
    :param protocol: the load balancer protocol.
    :param ip_address: the IP address of the load balancer
    :param client: the Octavia client
    :return: the compute node which hosts the Amphora.
    """
    amphorae = list_amphorae(load_balancer_id=load_balancer_id)
    amphora = get_master_amphora(amphorae=amphorae,
                                 port=port,
                                 protocol=protocol,
                                 ip_address=ip_address,
                                 client=client)
    server = nova.get_server(amphora['compute_id'])
    hostname = getattr(server, 'OS-EXT-SRV-ATTR:hypervisor_hostname')
    return topology.get_openstack_node(hostname=hostname)


# TODO (oschwart): use openstacksdk in this method implementation whenever
#  openstacksdk has its API call
def get_amphora_stats(amphora_id, client=None):
    """
    :param amphora_id: the amphora id
    :param client: The Octavia client
    :return (dict): The amphora stats dict.
    """

    return _client.octavia_client(client).amphora_stats_show(amphora_id)


def get_master_amphora(amphorae: typing.Iterable[AmphoraType],
                       port: int,
                       protocol: str,
                       ip_address: str,
                       client=None) -> AmphoraType:
    """Gets the master Amphora in a High Available LB
    (a loadbalancer which uses the Active/standby topology)

    :param amphorae: The list of amphoras (each represented by
     JSON).
    :param port: the load balancer port.
    :param protocol: the load balancer protocol.
    :param ip_address: the IP address of the load balancer
    :param client: the Octavia client
    :return amphora (dict): JSON of the Master Amphora.
    """

    amphorae = tobiko.select(amphorae)
    try:
        return amphorae.unique
    except tobiko.MultipleObjectsFound:
        # For a high available LB
        pass

    # Generate traffic on the LB so we can identify the current Master
    _validators.check_members_balanced(ip_address=ip_address,
                                       protocol=protocol,
                                       port=port,
                                       members_count=1,
                                       requests_count=1)

    # The amphora which has total_connections > 0 is the master.
    # Backup amphora will always have total_connections == 0.
    for amphora in amphorae:
        amphora_stats = get_amphora_stats(amphora_id=amphora['id'],
                                          client=client)
        for listener in list(amphora_stats.values())[0]:
            if listener['total_connections'] > 0:
                LOG.debug(f"Chosen amphora is {amphora['id']} with the "
                          f"following stats: {amphora_stats}")
                return amphora

    raise ValueError("Master Amphora wasn't found!")


def run_command_on_amphora(command: str,
                           lb_id: str = None,
                           lb_vip: str = None,
                           amp_id: str = None,
                           sudo: bool = False) -> str:
    """
    Run a given command on the master/single amphora

    :param command: The command to run on the amphora
    :param lb_id: The load balancer id whose amphora should run the command
    :param lb_vip: The loadbalancer VIP
    :param amp_id: The single/master amphora id
    :param sudo: (bool) Whether to run the command with sudo permissions
           on the amphora
    :return: The command output (str)
    """
    # Get the master/single amphora lb_network_ip
    if amp_id:
        amp_lb_network_ip = get_amphora(amp_id)['lb_network_ip']
    elif lb_id and lb_vip:
        amphorae = list_amphorae(load_balancer_id=lb_id)
        amphora = get_master_amphora(amphorae=amphorae,
                                     port=80,
                                     protocol='HTTP',
                                     ip_address=lb_vip)
        amp_lb_network_ip = amphora['lb_network_ip']
    else:
        raise ValueError('Either amphora id or both the loadbalancer id '
                         'and the loadbalancer floating ip need to be '
                         'provided.')

    # Find the undercloud ssh client and (any) controller ip
    def _get_overcloud_node_ssh_client(group):
        return topology.list_openstack_nodes(group=group)[0].ssh_client

    controller_ip = _get_overcloud_node_ssh_client('controller').host
    undercloud_client = _get_overcloud_node_ssh_client('undercloud')

    if not controller_ip or not undercloud_client:
        raise RuntimeError(f'Either controller ip {controller_ip} or '
                           f'undercloud ssh client {undercloud_client} was'
                           ' not found.')

    # Preparing ssh command
    osp_major_version = tripleo.overcloud_version().major
    if osp_major_version == 16:
        ssh_add_command = 'ssh-add'
    elif osp_major_version == 17:
        ssh_add_command = 'sudo -E ssh-add /etc/octavia/ssh/octavia_id_rsa'
    else:
        raise NotImplementedError('The ssh_add_command is not implemented '
                                  f'for OSP version {osp_major_version}.')

    ssh_agent_output = sh.execute(
        'ssh-agent -s',
        ssh_client=undercloud_client).stdout.strip()
    # Example: eval {ssh_agent_output} ssh-add
    start_agent_cmd = f'eval {ssh_agent_output} {ssh_add_command}; '

    # Example: ssh -A -t heat-admin@192.168.24.13
    controller_user = tripleo.get_overcloud_ssh_username()
    controller_ssh_command = f'ssh -A -t {controller_user}@{controller_ip}'

    amphora_user = CONF.tobiko.octavia.amphora_user
    # Example: ssh -o StrictHostKeyChecking=no cloud-user@172.24.0.214
    amphora_ssh_command = 'ssh -o StrictHostKeyChecking=no ' \
                          f'{amphora_user}@{amp_lb_network_ip}'
    full_amp_ssh_cmd = f'{controller_ssh_command} {amphora_ssh_command}'
    if sudo:
        command = f'sudo {command}'

    # Example:
    # $(ssh-agent -s) > ssh_agent_output
    #
    # eval {ssh_agent_output}; ssh-add; ssh -A -t heat-admin@192.168.24.13\
    # ssh -o StrictHostKeyChecking=no cloud-user@172.24.0.214 <command>
    command = f'{start_agent_cmd} {full_amp_ssh_cmd} {command}'
    out = sh.execute(command,
                     ssh_client=undercloud_client,
                     sudo=False).stdout.strip()

    # Removing the ssh-agent output
    # 'Agent pid 642546\n<output-we-want>' -> '<output-we-want>'
    formatted_out = '\n'.join(out.split('\n')[1:])
    return formatted_out



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.openstack.neutron import _agent
from tobiko.openstack.neutron import _client
from tobiko.openstack.neutron import _extension
from tobiko.openstack.neutron import _floating_ip
from tobiko.openstack.neutron import _port
from tobiko.openstack.neutron import _quota_set
from tobiko.openstack.neutron import _network
from tobiko.openstack.neutron import _router
from tobiko.openstack.neutron import _security_group
from tobiko.openstack.neutron import _subnet
from tobiko.openstack.neutron import _subnet_pool


SERVER = 'neutron-server'
METADATA_IPv4 = '169.254.169.254'
DHCP_AGENT = _agent.DHCP_AGENT
L3_AGENT = _agent.L3_AGENT
METADATA_AGENT = _agent.METADATA_AGENT
OPENVSWITCH_AGENT = _agent.OPENVSWITCH_AGENT
OVN_CONTROLLER = _agent.OVN_CONTROLLER
OVN_METADATA_AGENT = _agent.OVN_METADATA_AGENT
NEUTRON_OVN_METADATA_AGENT = _agent.NEUTRON_OVN_METADATA_AGENT
OVN_BGP_AGENT = _agent.OVN_BGP_AGENT
FRR = _agent.FRR
NEUTRON = _agent.NEUTRON
DEFAULT_SG_NAME = _security_group.DEFAULT_SG_NAME
STATEFUL_OVN_ACTION = _security_group.STATEFUL_OVN_ACTION
STATELESS_OVN_ACTION = _security_group.STATELESS_OVN_ACTION

AgentNotFoundOnHost = _agent.AgentNotFoundOnHost
NotFound = _client.NotFound
NeutronAgentType = _agent.NeutronAgentType
SecurityGroupType = _security_group.SecurityGroupType
SecurityGroupIdOrNameType = _security_group.SecurityGroupIdOrNameType
get_l3_agent_mode = _agent.get_l3_agent_mode
find_l3_agent_hosting_router = _agent.find_l3_agent_hosting_router
list_agents = _agent.list_agents
list_dhcp_agent_hosting_network = _agent.list_dhcp_agent_hosting_network
list_l3_agent_hosting_routers = _agent.list_l3_agent_hosting_routers
list_networking_agents = _agent.list_networking_agents
skip_if_missing_networking_agents = _agent.skip_if_missing_networking_agents
skip_unless_is_ovn = _agent.skip_unless_is_ovn
skip_unless_is_ovs = _agent.skip_unless_is_ovs
skip_if_is_old_ovn = _agent.skip_if_is_old_ovn
has_ovn = _agent.has_ovn
has_ovs = _agent.has_ovs

NeutronClientFixture = _client.NeutronClientFixture
ServiceUnavailable = _client.ServiceUnavailable
NeutronClient = _client.NeutronClient
NeutronClientException = _client.NeutronClientException
NeutronClientType = _client.NeutronClientType
neutron_client = _client.neutron_client
get_neutron_client = _client.get_neutron_client

get_networking_extensions = _extension.get_networking_extensions
missing_networking_extensions = _extension.missing_networking_extensions
has_networking_extensions = _extension.has_networking_extensions
skip_if_missing_networking_extensions = (
    _extension.skip_if_missing_networking_extensions)

create_floating_ip = _floating_ip.create_floating_ip
delete_floating_ip = _floating_ip.delete_floating_ip
get_floating_ip = _floating_ip.get_floating_ip
get_floating_ip_id = _floating_ip.get_floating_ip_id
find_floating_ip = _floating_ip.find_floating_ip
list_floating_ips = _floating_ip.list_floating_ips
update_floating_ip = _floating_ip.update_floating_ip
FloatingIpType = _floating_ip.FloatingIpType
FloatingIpIdType = _floating_ip.FloatingIpIdType
NoSuchFloatingIp = _floating_ip.NoSuchFloatingIp

create_port = _port.create_port
delete_port = _port.delete_port
get_port = _port.get_port
get_port_id = _port.get_port_id
find_device_ip_address = _port.find_device_ip_address
find_port = _port.find_port
find_port_ip_address = _port.find_port_ip_address
list_ports = _port.list_ports
list_port_ip_addresses = _port.list_port_ip_addresses
list_device_ip_addresses = _port.list_device_ip_addresses
update_port = _port.update_port
PortType = _port.PortType
PortIdType = _port.PortIdType
NoSuchPort = _port.NoSuchPort

get_neutron_quota_set = _quota_set.get_neutron_quota_set
set_neutron_quota_set = _quota_set.set_neutron_quota_set
ensure_neutron_quota_limits = _quota_set.ensure_neutron_quota_limits
EnsureNeutronQuotaLimitsError = _quota_set.EnsureNeutronQuotaLimitsError

create_network = _network.create_network
delete_network = _network.delete_network
get_network = _network.get_network
get_network_id = _network.get_network_id
find_network = _network.find_network
list_networks = _network.list_networks
list_network_nameservers = _network.list_network_nameservers
NoSuchNetwork = _network.NoSuchNetwork
NetworkType = _network.NetworkType
NetworkIdType = _network.NetworkIdType

add_router_interface = _router.add_router_interface
create_router = _router.create_router
delete_router = _router.delete_router
get_ovs_router_namespace = _router.get_ovs_router_namespace
get_router = _router.get_router
get_router_id = _router.get_router_id
remove_router_interface = _router.remove_router_interface
wait_for_master_and_backup_agents = _router.wait_for_master_and_backup_agents
update_router = _router.update_router
RouterType = _router.RouterType
RouterIdType = _router.RouterIdType
NoSuchRouter = _router.NoSuchRouter

create_subnet = _subnet.create_subnet
delete_subnet = _subnet.delete_subnet
ensure_subnet_gateway = _subnet.ensure_subnet_gateway
get_subnet = _subnet.get_subnet
get_subnet_id = _subnet.get_subnet_id
find_subnet = _subnet.find_subnet
list_subnets = _subnet.list_subnets
SubnetType = _subnet.SubnetType
SubnetIdType = _subnet.SubnetIdType
NoSuchSubnet = _subnet.NoSuchSubnet

SubnetPoolType = _subnet_pool.SubnetPoolType
SubnetPoolIdType = _subnet_pool.SubnetPoolIdType
NoSuchSubnetPool = _subnet_pool.NoSuchSubnetPool
get_subnet_pool = _subnet_pool.get_subnet_pool
create_subnet_pool = _subnet_pool.create_subnet_pool
delete_subnet_pool = _subnet_pool.delete_subnet_pool
find_subnet_pool = _subnet_pool.find_subnet_pool
list_subnet_pools = _subnet_pool.list_subnet_pools

list_security_groups = _security_group.list_security_groups
get_security_group = _security_group.get_security_group
get_default_security_group = _security_group.get_default_security_group
create_security_group = _security_group.create_security_group
update_security_group = _security_group.update_security_group
delete_security_group = _security_group.delete_security_group
create_security_group_rule = _security_group.create_security_group_rule



# Copyright 2023 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.podified import _topology
from tobiko.podified import _openshift
from tobiko.podified import containers


EDPM_NODE = _topology.EDPM_NODE
OCP_WORKER = _topology.OCP_WORKER
EDPM_COMPUTE_GROUP = _openshift.EDPM_COMPUTE_GROUP
EDPM_NETWORKER_GROUP = _openshift.EDPM_NETWORKER_GROUP
EDPM_OTHER_GROUP = _openshift.EDPM_OTHER_GROUP

PodifiedTopology = _topology.PodifiedTopology

skip_if_not_podified = _topology.skip_if_not_podified
skip_if_podified = _topology.skip_if_podified

get_dataplane_ssh_keypair = _openshift.get_dataplane_ssh_keypair
has_podified_cp = _openshift.has_podified_cp
get_ovndbcluter = _openshift.get_ovndbcluter

get_container_runtime_name = containers.get_container_runtime_name



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

import tobiko
from tobiko import shiftstack


@shiftstack.skip_unless_has_shiftstack()
class ShiftstackNovaTest(testtools.TestCase):

    def test_list_shifstack_nodes(self):
        nodes = shiftstack.list_shiftstack_nodes()
        self.assertIsInstance(nodes, tobiko.Selection)
        self.assertNotEqual([], nodes)
        return nodes

    def test_find_shifstack_nodes(self):
        nodes = self.test_list_shifstack_nodes()
        node = shiftstack.find_shiftstack_node(name=nodes.first.name)
        self.assertEqual(nodes.first.id, node.id)



# Copyright 2021 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

from keystoneclient import base
from keystoneclient.v2_0 import tenants as tenants_v2
from keystoneclient.v2_0 import users as users_v2
from keystoneclient.v3 import projects as projects_v3
from keystoneclient.v3 import users as users_v3


KeystoneResourceType = typing.Union[str, base.Resource]
ProjectType = typing.Union[str, tenants_v2.Tenant, projects_v3.Project]
UserType = typing.Union[str, users_v2.User, users_v3.User]


def get_project_id(
        project: typing.Optional[ProjectType] = None,
        session=None) -> str:
    if project is not None:
        return get_keystone_resource_id(project)
    if session is not None:
        return session.get_project_id()
    raise ValueError("'project' and 'session' can't be None ata the same "
                     "time.")


def get_user_id(
        user: typing.Optional[UserType] = None,
        session=None) -> str:
    if user is not None:
        return get_keystone_resource_id(user)
    if session is not None:
        return session.get_user_id()
    raise ValueError("'project' and 'session' can't be None ata the same "
                     "time.")


def get_keystone_resource_id(resource: KeystoneResourceType) -> str:
    if isinstance(resource, str):
        return resource

    if isinstance(resource, base.Resource):
        return resource.id

    raise TypeError(f"Object {resource} is not a valid Keystone resource "
                    "type")



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import typing

import netaddr
from oslo_log import log


LOG = log.getLogger(__name__)

PingAddressType = typing.Union[str, netaddr.IPAddress]


class PingParameters(collections.namedtuple('PingParameters',
                                            ['host',
                                             'count',
                                             'deadline',
                                             'fragmentation',
                                             'interval',
                                             'ip_version',
                                             'packet_size',
                                             'source',
                                             'timeout',
                                             'network_namespace'])):
    """Recollect parameters to be used to format ping command line

    PingParameters class is a data model recollecting parameters used to
    create a ping command line. It provides the feature of copying default
    values from another instance of PingParameters passed using constructor
    parameter 'default'.
    """


PING_PARAMETERS_NAMES = PingParameters._fields


def get_ping_parameters(default=None, **ping_params):
    """Get ping parameters eventually merging them with given extra parameters

    The only difference with init_parameters function is that in case
    default parameter is not None not any extra parameters is given, then
    it simply return given default instance, without performing any validation.
    """
    if default and not ping_params:
        return default
    else:
        return ping_parameters(default=default, **ping_params)


def ping_parameters(default=None, count=None, deadline=None,
                    fragmentation=None,
                    host: typing.Optional[PingAddressType] = None,
                    interval=None, ip_version=None, packet_size=None,
                    source=None, timeout=None, network_namespace=None):
    """Validate parameters and initialize a new PingParameters instance

    :param default: (PingParameters or None) instance from where to take
        default values when other value is not provided. If None (that is the
        default value) it will copy default parameters from
        DEFAULT_PING_PARAMETERS
    :param count: (int or None) number of ping ICMP message expecting to be
        received before having a success. Default value can be configured using
        'count' option in [ping] config section.
    :param host: (str or None) IP address or host name to send ICMP
        messages to. It is required to format a valid ping command, therefore
        no default value exists for this parameter.
    :param deadline: (int or None) positive number representing the maximum
        number of seconds ping command can send ICMP messages before stop
        executing. Default value can be configured using 'deadline' option in
        [ping] config section.
    :param fragmentation: (bool or None) when False this would tell ping
        to forbid ICMP messages fragmentation. Default value can be configured
        using 'fragmentation' option in [ping] config section. Fragmentation
        can't be disabled when using ping provided by BusyBox (IE with CirrOS
        images).
    :param interval: (int or None) interval of time before sending following
        ICMP message. Default value can be configured using 'interval' option
        in [ping] config section.
    :param ip_version: (4, 6 or None) If not None it makes sure it will
        use specified IP version for sending ICMP packages.
    :param packet_size: (int or None) if not None, it specifies the total ICMP
        message size (headers + payload).
    :param source: (str or None) IP address or interface name from where
        to send ICMP message.
    :param timeout: (int or None) time in seconds after which ping operation
        would raise PingFailed exception.
    :raises TypeError: in case some parameter cannot be converted to right
        expected type
    :raises ValueError: in case some parameter has an unexpected value
    """

    if default is None:
        default = default_ping_parameters()

    return PingParameters(
        count=get_positive_integer('count', count, default),
        host=get_address('host', host, default),
        deadline=get_positive_integer('deadline', deadline, default),
        fragmentation=get_boolean('fragmentation', fragmentation, default),
        interval=get_positive_integer('interval', interval, default),
        ip_version=get_positive_integer('ip_version', ip_version, default),
        packet_size=get_positive_integer('packet_size', packet_size, default),
        source=get_address('source', source, default),
        timeout=get_positive_integer('timeout', timeout, default),
        network_namespace=get_string('network_namespace', network_namespace,
                                     default))


def default_ping_parameters():
    from tobiko import config
    CONF = config.CONF
    return ping_parameters(default=False,
                           count=CONF.tobiko.ping.count,
                           deadline=CONF.tobiko.ping.deadline,
                           fragmentation=CONF.tobiko.ping.fragmentation,
                           interval=CONF.tobiko.ping.interval,
                           packet_size=CONF.tobiko.ping.packet_size,
                           timeout=CONF.tobiko.ping.timeout)


def get_ping_ip_version(parameters):
    ip_version = parameters.ip_version
    if ip_version is not None:
        ip_version = int(ip_version)
        if ip_version not in [4, 6]:
            message = "Invalid IP version: {!r}".format(ip_version)
            raise ValueError(message)
    for address in [parameters.host, parameters.source]:
        if isinstance(address, netaddr.IPAddress):
            if ip_version != address.version:
                if ip_version:
                    meassage = ("{!s} address IP version is not {!r}"
                                ).format(address, ip_version)
                    raise ValueError(meassage)
                ip_version = address.version
    return ip_version


def get_ping_payload_size(parameters):
    packet_size = parameters.packet_size
    if packet_size is None:
        return None

    header_size = get_ping_header_size(parameters)
    if packet_size < header_size:
        message = ("packet size {packet_size!s} can't be smaller than "
                   "header size {header_size!s}").format(
                       packet_size=packet_size,
                       header_size=header_size)
        raise ValueError(message)
    return packet_size - header_size


def get_positive_integer(name, value, default=None):
    if value is None and default:
        return get_positive_integer(name, getattr(default, name))
    if value is not None:
        value = int(value)
        if value <= 0:
            message = "{!r} value must be positive: {!r}".format(
                name, value)
            raise ValueError(message)
    return value


def get_boolean(name, value, default=None):
    if value is None and default:
        return get_boolean(name, getattr(default, name))
    if value is not None:
        value = bool(value)
    return value


def get_address(name: str, value: typing.Optional[PingAddressType],
                default=None) -> typing.Optional[PingAddressType]:
    if value is None:
        if default:
            return get_address(name, getattr(default, name))
        else:
            return None
    if isinstance(value, netaddr.IPAddress):
        return value
    elif isinstance(value, str):
        try:
            return netaddr.IPAddress(value)
        except netaddr.core.AddrFormatError:
            # NOTE: value may be an host name so this is fine
            return value
    else:
        raise TypeError(f"Object '{value}' is not a valid address")


def get_string(name, value, default=None):
    if value is None and default:
        return get_string(name, getattr(default, name))
    if value is not None:
        value = str(value)
    return value


IP_HEADER_SIZE = {4: 20, 6: 40}
ICMP_HEADER_SIZE = {4: 8, 6: 4}


def get_ping_header_size(parameters):
    ip_version = get_ping_ip_version(parameters)
    if ip_version is None:
        message = "can't get ICMP header size without knowing IP version"
        raise ValueError(message)

    if ip_version not in IP_HEADER_SIZE or ip_version not in ICMP_HEADER_SIZE:
        message = "Invalid IP version: {!r}".format(ip_version)
        raise ValueError(message)

    return IP_HEADER_SIZE[ip_version] + ICMP_HEADER_SIZE[ip_version]



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

from octaviaclient.api.v2 import octavia

import tobiko
from tobiko.openstack import _client, openstacksdkclient
from tobiko.openstack import keystone


OCTAVIA_CLIENT_CLASSSES = octavia.OctaviaAPI,


OctaviaClientType = typing.Union[octavia.OctaviaAPI,
                                 'OctaviaClientFixture']


def get_octavia_endpoint(keystone_client=None):
    return keystone.find_service_endpoint(name='octavia',
                                          client=keystone_client)


class OctaviaClientFixture(_client.OpenstackClientFixture):

    def init_client(self, session):
        keystone_client = keystone.get_keystone_client(session=session)
        endpoint = get_octavia_endpoint(keystone_client=keystone_client)
        return octavia.OctaviaAPI(session=session, endpoint=endpoint.url)


class OctaviaClientManager(_client.OpenstackClientManager):

    def create_client(self, session):
        return OctaviaClientFixture(session=session)


CLIENTS = OctaviaClientManager()


@keystone.skip_if_missing_service(name='octavia')
def octavia_client(obj: OctaviaClientType = None) -> octavia.OctaviaAPI:
    if obj is None:
        return get_octavia_client()

    if isinstance(obj, OCTAVIA_CLIENT_CLASSSES):
        return obj

    fixture = tobiko.setup_fixture(obj)
    if isinstance(fixture, OctaviaClientFixture):
        return fixture.client

    message = "Object {!r} is not an OctaviaClientFixture".format(obj)
    raise TypeError(message)


def get_octavia_client(session=None, shared=True, init_client=None,
                       manager=None):
    manager = manager or CLIENTS
    client = manager.get_client(session=session, shared=shared,
                                init_client=init_client)
    tobiko.setup_fixture(client)
    return client.client


def list_members(pool_id: str):
    os_sdk_client = openstacksdkclient.openstacksdk_client()
    return os_sdk_client.load_balancer.members(pool=pool_id)


def list_load_balancers(**lb_kwargs):
    os_sdk_client = openstacksdkclient.openstacksdk_client()
    return os_sdk_client.load_balancer.load_balancers(**lb_kwargs)


def find_load_balancer(lb_name: str):
    os_sdk_client = openstacksdkclient.openstacksdk_client()
    return os_sdk_client.load_balancer.find_load_balancer(lb_name)


def create_load_balancer(lb_kwargs):
    os_sdk_client = openstacksdkclient.openstacksdk_client()
    return os_sdk_client.load_balancer.create_load_balancer(**lb_kwargs)


def find_listener(listener_name: str):
    os_sdk_client = openstacksdkclient.openstacksdk_client()
    return os_sdk_client.load_balancer.find_listener(listener_name)


def create_listener(listener_kwargs):
    os_sdk_client = openstacksdkclient.openstacksdk_client()
    return os_sdk_client.load_balancer.create_listener(**listener_kwargs)


def find_pool(pool_name: str):
    os_sdk_client = openstacksdkclient.openstacksdk_client()
    return os_sdk_client.load_balancer.find_pool(pool_name)


def create_pool(pool_kwargs):
    os_sdk_client = openstacksdkclient.openstacksdk_client()
    return os_sdk_client.load_balancer.create_pool(**pool_kwargs)


def find_member(member_name: str, pool: str):
    # Note that pool could be either id or name
    os_sdk_client = openstacksdkclient.openstacksdk_client()
    return os_sdk_client.load_balancer.find_member(member_name, pool)


def create_member(member_kwargs):
    os_sdk_client = openstacksdkclient.openstacksdk_client()
    return os_sdk_client.load_balancer.create_member(**member_kwargs)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from collections import abc

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack.neutron import _client


@keystone.skip_unless_has_keystone_credentials()
class NetworkingExtensionsFixture(tobiko.SharedFixture):

    client = None
    extensions = None

    def setup_fixture(self):
        self.setup_client()
        self.get_networking_extensions()

    def setup_client(self):
        self.client = _client.get_neutron_client()

    def get_networking_extensions(self):
        extensions = self.client.list_extensions()
        if isinstance(extensions, abc.Mapping):
            extensions = extensions['extensions']
        ignore_extensions = set(
            tobiko.tobiko_config().neutron.ignore_extensions)
        self.extensions = frozenset(e['alias']
                                    for e in extensions
                                    if e['alias'] not in ignore_extensions)


def get_networking_extensions():
    return tobiko.setup_fixture(NetworkingExtensionsFixture).extensions


def missing_networking_extensions(*extensions):
    return sorted(frozenset(extensions) - get_networking_extensions())


def has_networking_extensions(*extensions):
    return not missing_networking_extensions(*extensions)


def skip_if_missing_networking_extensions(*extensions):
    return tobiko.skip_if('missing networking extensions: {return_value!r}',
                          missing_networking_extensions, *extensions)


if __name__ == '__main__':
    import sys
    sys.stdout.write('\n'.join(sorted(get_networking_extensions())))
    sys.stdout.flush()



# Copyright 2021 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import tobiko
from tobiko.openstack.topology import _topology
from tobiko.shell import sh


def list_nodes_processes(command: str = None,
                         command_line: sh.ShellCommandType = None,
                         hostnames: typing.Iterable[str] = None,
                         topology: _topology.OpenStackTopology = None,
                         **list_processes_params) \
        -> tobiko.Selection[sh.PsProcess]:
    processes = tobiko.Selection[sh.PsProcess]()
    nodes = _topology.list_openstack_nodes(topology=topology,
                                           hostnames=hostnames)
    for node in nodes:
        processes += sh.list_processes(command=command,
                                       command_line=command_line,
                                       ssh_client=node.ssh_client,
                                       **list_processes_params)
    return processes



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing
from collections import abc

from oslo_log import log
from designateclient.v2 import client

import tobiko
from tobiko.openstack import _client

DESIGNATE_CLIENT_CLASSES = client.Client,

LOG = log.getLogger(__name__)


class DesignateClientFixture(_client.OpenstackClientFixture):

    def init_client(self, session) -> client.Client:
        return client.Client(session=session)


class DesignateClientManager(_client.OpenstackClientManager):

    def create_client(self, session) -> DesignateClientFixture:
        return DesignateClientFixture(session=session)


CLIENTS = DesignateClientManager()

DesignateClientType = typing.Union[client.Client, DesignateClientFixture]


DesignateZone = typing.Mapping[str, typing.Any]
DesignateZoneType = typing.Union[str, typing.Mapping[str, typing.Any]]


def designate_client(obj: DesignateClientType = None):
    if obj is None:
        return get_designate_client()

    if isinstance(obj, client.Client):
        return obj

    fixture = tobiko.setup_fixture(obj)
    if isinstance(fixture, DesignateClientFixture):
        return fixture.client

    message = "Object {!r} is not an DesignateClientFixture".format(obj)
    raise TypeError(message)


def get_designate_client(session=None,
                         shared=True,
                         init_client=None,
                         manager: DesignateClientManager = None) \
        -> client.Client:
    manager = manager or CLIENTS
    fixture = manager.get_client(session=session,
                                 shared=shared,
                                 init_client=init_client)
    tobiko.setup_fixture(fixture)
    return fixture.client


def designate_zone_id(zone: DesignateZoneType) -> str:
    if isinstance(zone, str):
        return zone
    elif isinstance(zone, abc.Mapping):
        return zone['id']
    else:
        raise TypeError(f'{zone} object is an invalid Designate zone type')


def get_designate_zone(zone: str) -> DesignateZone:
    zone_id = designate_zone_id(zone)
    return designate_client().zones.get(zone_id)


def list_recordsets(zone: str) -> DesignateZone:
    zone_id = designate_zone_id(zone)
    return designate_client().recordsets.list(zone_id)


def create_recordsets(zone: str, name: str, type_: str,
                      records: typing.List[str]):
    zone_id = designate_zone_id(zone)
    return designate_client().recordsets.create(
        zone_id, name=name, type_=type_, records=records)


def get_recordset(zone_id: str, recordset_id: str):
    return designate_client().recordsets.get(
        zone=zone_id, recordset=recordset_id)


def wait_for_status(status_key, status, get_client, object_id,
                    interval: tobiko.Seconds = None,
                    timeout: tobiko.Seconds = None,
                    **kwargs):
    """Waits for an object to reach a specific status.

    :param status_key: The key of the status field in the response.
                       Ex. status
    :param status: The status to wait for. Ex. "ACTIVE"
    :param get_client: The tobiko client get method.
                        Ex. _client.get_zone
    :param object_id: The id of the object to query.
    :param interval: How often to check the status, in seconds.
    :param timeout: The maximum time, in seconds, to check the status.
    :raises TimeoutException: The object did not achieve the status or ERROR in
                              the check_timeout period.
    :raises UnexpectedStatusException: The request returned an unexpected
                                       response code.
    """
    for attempt in tobiko.retry(timeout=timeout,
                                interval=interval,
                                default_timeout=300.,
                                default_interval=5.):
        response = get_client(object_id, **kwargs)
        if response[status_key] == status:
            return response

        attempt.check_limits()

        LOG.debug(f"Waiting for {get_client.__name__} {status_key} to get "
                  f"from '{response[status_key]}' to '{status}'...")



from __future__ import absolute_import

import enum
import io
import time

from oslo_log import log
import pandas

import tobiko
from tobiko import config
from tobiko.tripleo import overcloud
from tobiko.shell import sh
from tobiko.shell import ssh
from tobiko.openstack import topology


CONF = config.CONF
LOG = log.getLogger(__name__)

GALERA_RESOURCE = "galera-bundle"
HAPROXY_RESOURCE = "haproxy-bundle"
OVN_DBS_RESOURCE = "ovn-dbs-bundle"


class PcsResourceException(tobiko.TobikoException):
    message = "pcs cluster is not in a healthy state"


def get_pcs_resources_table(timeout=720, interval=2) -> pandas.DataFrame:
    """
    get pcs status from a controller and parse it
    to have it's resources states in check
       returns :
       rabbitmq-bundle-0    (ocf::heartbeat:rabbitmq-cluster):      Started con
       troller-0
     ip-10.0.0.101  (ocf::heartbeat:IPaddr2):       Started controller-1
       openstack-cinder-volume-docker-0     (ocf::heartbeat:docker):        Sta
       rted controller-0

    :return: dataframe of pcs resources stats table
    """
    # prevent pcs table read failure while pacemaker is starting
    for attempt in tobiko.retry(timeout=timeout,
                                interval=interval):
        try:
            output = run_pcs_status(options=['resources'], grep_str='ocf')
            # remove the first column when it only includes '*' characters
            output = output.replace('*', '').strip()
            stream = io.StringIO(output)
            table: pandas.DataFrame = pandas.read_csv(
                stream, delim_whitespace=True, header=None)
            table.columns = ['resource', 'resource_type', 'resource_state',
                             'overcloud_node']
        except (ValueError, sh.ShellCommandFailed, sh.ShellTimeoutExpired):
            if attempt.is_last:
                raise
            LOG.exception('Failed to obtain pcs status table - Retrying...')
        else:
            break

    LOG.debug("Got pcs status :\n%s", table)
    return table


def get_pcs_prefix_and_status_values():
    if topology.verify_osp_version('17.0', lower=True):
        ocf_prefix = "ocf::"
        promoted_status_str = "Master"
        unpromoted_status_str = "Slave"
    else:
        ocf_prefix = "ocf:"
        promoted_status_str = "Promoted"
        unpromoted_status_str = "Unpromoted"
    return ocf_prefix, promoted_status_str, unpromoted_status_str


class PacemakerResourcesStatus(object):
    """
    class to handle pcs resources checks
    """
    def __init__(self):
        self.pcs_df = get_pcs_resources_table()
        (self.ocf_prefix,
         self.promoted_status_str,
         self.unpromoted_status_str) = get_pcs_prefix_and_status_values()

    def container_runtime(self):

        if not self.pcs_df[(self.pcs_df['resource_type'] ==
                            f"({self.ocf_prefix}heartbeat:docker):")].empty:
            return 'docker'
        if not self.pcs_df[(self.pcs_df['resource_type'] ==
                            f"({self.ocf_prefix}heartbeat:podman):")].empty:
            return 'podman'

    def resource_count(self, resource_type):
        return self.pcs_df[(self.pcs_df['resource_type'] == resource_type)][
            'resource_state'].count()

    def resource_count_in_state(self, resource_type, resource_state):
        return self.pcs_df[(self.pcs_df['resource_type'] ==
                            resource_type) & (self.pcs_df['resource_state'] ==
                                              resource_state)][
            'resource_state'].count()

    def rabbitmq_resource_healthy(self):
        rabbitmq_resource_str = \
            f"({self.ocf_prefix}heartbeat:rabbitmq-cluster):"
        nodes_num = self.resource_count(rabbitmq_resource_str)
        started_num = self.resource_count_in_state(
            rabbitmq_resource_str, "Started")
        if nodes_num == started_num and nodes_num > 0:
            LOG.info("pcs status check: resource rabbitmq is in healthy state")
            return True
        else:
            LOG.info("pcs status check: resource rabbitmq not in healthy "
                     "state")
            return False

    def galera_resource_healthy(self):
        galera_resource_str = f"({self.ocf_prefix}heartbeat:galera):"
        nodes_num = self.resource_count(galera_resource_str)
        master_num = self.resource_count_in_state(
            galera_resource_str, self.promoted_status_str)
        if nodes_num == master_num and nodes_num > 0:
            LOG.info("pcs status check: resource galera is in healthy state")
            return True
        else:
            LOG.info("pcs status check: resource galera not in healthy state")
            return False

    def redis_resource_healthy(self):
        redis_resource_str = f"({self.ocf_prefix}heartbeat:redis):"
        if not overcloud.is_redis_expected():
            LOG.info("redis resource not expected on OSP 17 "
                     "and later releases by default")
            return self.pcs_df.query(
                f'resource_type == "{redis_resource_str}"').empty
        nodes_num = self.resource_count(redis_resource_str)
        master_num = self.resource_count_in_state(
            redis_resource_str, self.promoted_status_str)
        slave_num = self.resource_count_in_state(
            redis_resource_str, self.unpromoted_status_str)
        if (master_num == 1) and (slave_num == nodes_num - master_num):
            LOG.info("pcs status check: resource redis is in healthy state")
            return True
        else:
            LOG.info("pcs status check: resource redis not in healthy state")
            return False

    def vips_resource_healthy(self):
        if CONF.tobiko.rhosp.has_external_load_balancer:
            LOG.info("external load balancer used - "
                     "we can skip vips_resource sanity")
            return True
        else:
            vips_resource_str = f"({self.ocf_prefix}heartbeat:IPaddr2):"
            nodes_num = self.resource_count(vips_resource_str)
            started_num = self.resource_count_in_state(
                vips_resource_str, "Started")
            if nodes_num == started_num and nodes_num > 0:
                LOG.info("pcs status check: resources vips are "
                         "in healthy state")
                return True
            else:
                LOG.info(
                    "pcs status check: resources"
                    " vips are not in healthy state")
                return False

    def ha_proxy_cinder_healthy(self):
        if CONF.tobiko.rhosp.has_external_load_balancer:
            LOG.info("external load balancer used "
                     "- we can skip ha_proxy_resource sanity")
            return True
        else:
            ha_proxy_resource_str = (f"({self.ocf_prefix}heartbeat:"
                                     f"{self.container_runtime()}):")
            nodes_num = self.resource_count(ha_proxy_resource_str)
            started_num = self.resource_count_in_state(
                ha_proxy_resource_str, "Started")
            if nodes_num == started_num and nodes_num > 0:
                LOG.info("pcs status check: resources ha_proxy and"
                         " cinder are in healthy state")
                return True
            else:
                LOG.info(
                    "pcs status check: resources ha_proxy and cinder "
                    "are not in healthy state")
                return False

    def ovn_resource_healthy(self):
        ovn_resource_str = f"({self.ocf_prefix}ovn:ovndb-servers):"
        if self.pcs_df.query(
                f'resource_type == "{ovn_resource_str}"').empty:
            LOG.info('pcs status check: ovn is not deployed, skipping ovn '
                     'resource check')
            return True
        nodes_num = self.resource_count(ovn_resource_str)
        master_num = self.resource_count_in_state(
            ovn_resource_str, self.promoted_status_str)
        slave_num = self.resource_count_in_state(
            ovn_resource_str, self.unpromoted_status_str)
        if (master_num == 1) and (slave_num == nodes_num - master_num):
            LOG.info(
                "pcs status check: resource ovn is in healthy state")
            return True
        else:
            LOG.info(
                "pcs status check: resource ovn is in not in "
                "healthy state")
            return False

    @property
    def all_healthy(self):
        """
        check if each resource is in healthy order
        and return a global healthy status
        :return: Bool
        """
        for attempt_number in range(360):

            try:

                if all([
                   self.rabbitmq_resource_healthy(),
                   self.galera_resource_healthy(),
                   self.redis_resource_healthy(),
                   self.vips_resource_healthy(),
                   self.ha_proxy_cinder_healthy(),
                   self.ovn_resource_healthy()
                   ]):
                    LOG.info("pcs status checks: all resources are"
                             " in healthy state")
                    return True
                else:

                    LOG.info("pcs status check: not all resources are "
                             "in healthy "
                             "state")
                    raise PcsResourceException()
            except PcsResourceException:
                # reread pcs status
                LOG.info('Retrying pacemaker resource checks attempt '
                         '{} of 360'.format(attempt_number))
                time.sleep(1)
                self.pcs_df = get_pcs_resources_table()
        # exhausted all retries
        tobiko.fail('pcs cluster is not in a healthy state')


def get_overcloud_nodes_running_pcs_resource(resource=None,
                                             resource_type=None,
                                             resource_state=None):
    """
    Check what nodes are running the specified resource/type/state
    resource/type/state: exact str of a resource name as seen in pcs status
    :return: list of overcloud nodes
    """
    # pylint: disable=no-member
    pcs_df = get_pcs_resources_table()
    if resource:
        pcs_df_query_resource = pcs_df.query('resource=="{}"'.format(
                                        resource))
        return pcs_df_query_resource['overcloud_node'].unique().tolist()

    if resource_type and resource_state:
        pcs_df_query_resource_type_state = pcs_df.query(
            'resource_type=="{}" and resource_state=="{}"'.format(
                resource_type, resource_state))
        return pcs_df_query_resource_type_state[
            'overcloud_node'].unique().tolist()

    if resource_type and not resource_state:
        pcs_df_query_resource_type = pcs_df.query(
            'resource_type=="{}"'.format(resource_type))
        return pcs_df_query_resource_type['overcloud_node'].unique().tolist()


def get_resource_master_node(resource_type=None):
    get_overcloud_nodes_running_pcs_resource(
        resource_type=resource_type, resource_state='Master')


def get_ovn_db_master_node():
    ocf_prefix, promoted_status_str, _ = get_pcs_prefix_and_status_values()
    return get_overcloud_nodes_running_pcs_resource(
        resource_type=f'({ocf_prefix}ovn:ovndb-servers):',
        resource_state=promoted_status_str)


def get_overcloud_resource(resource_type=None,
                           resource_state=None):
    """
    Check what nodes are running the specified resource/type/state
    resource/type/state: exact str of a resource name as seen in pcs status
    :return: list of overcloud nodes
    """
    pcs_df = get_pcs_resources_table()

    if resource_type and resource_state:
        pcs_df_query_resource_type_state = pcs_df.query(
            'resource_type=="{}" and resource_state=="{}"'.format(
                resource_type, resource_state))
        return pcs_df_query_resource_type_state[
            'resource'].unique().tolist()

    if resource_type and not resource_state:
        # pylint: disable=no-member
        pcs_df_query_resource_type = pcs_df.query(
            'resource_type=="{}"'.format(resource_type))
        return pcs_df_query_resource_type['resource'].unique().tolist()


def instanceha_deployed():
    """check IHA deployment
    checks for existence of the nova-evacuate resource"""
    if overcloud.has_overcloud():
        return get_overcloud_nodes_running_pcs_resource(
            resource='nova-evacuate')
    else:
        return False


skip_if_instanceha_not_delpoyed = tobiko.skip_unless(
    'instanceha not delpoyed', instanceha_deployed)


def fencing_deployed():
    """check fencing deployment
    checks for existence of the stonith-fence type resources"""
    fencing_output = run_pcs_status(grep_str="stonith:fence_ipmilan")

    if fencing_output:
        return True
    else:
        return False


skip_if_fencing_not_deployed = tobiko.skip_unless(
    'fencing not delpoyed', fencing_deployed)


def run_pcs_status(ssh_client: ssh.SSHClientFixture = None,
                   options: list = None,
                   grep_str: str = None) -> str:
    command_args = ['status']
    command_args += options or []

    output = execute_pcs(command_args,
                         ssh_client=ssh_client,
                         sudo=True)

    if not grep_str:
        return output

    output_ocf_lines = []
    for line in output.splitlines():
        if grep_str in line:
            output_ocf_lines.append(line)

    return '\n'.join(output_ocf_lines)


class PcsResourceOperation(enum.Enum):
    DISABLE = "disable"
    ENABLE = "enable"
    RESTART = "restart"
    SHOW = "show"

    def __init__(self, pcsoperation: str):
        self.pcsoperation = pcsoperation


DISABLE = PcsResourceOperation.DISABLE
ENABLE = PcsResourceOperation.ENABLE
RESTART = PcsResourceOperation.RESTART
SHOW = PcsResourceOperation.SHOW


def run_pcs_resource_operation(resource: str,
                               operation: PcsResourceOperation,
                               ssh_client: ssh.SSHClientFixture = None,
                               node: str = None,
                               operation_wait: int = 60,
                               retry_timeout: float = 180.,
                               retry_interval: float = 5.) -> str:
    tobiko.check_valid_type(operation, PcsResourceOperation)

    command_args = ['resource', operation.pcsoperation, resource]
    if node is not None:
        command_args.append(node)

    command_args.append(f'--wait={operation_wait}')

    # add stderr to the output if the operation is disable or enable
    add_stderr = operation in (DISABLE, ENABLE)
    # execute the command with retries
    for attempt in tobiko.retry(timeout=retry_timeout,
                                interval=retry_interval):
        try:
            output = execute_pcs(command_args,
                                 ssh_client=ssh_client,
                                 add_stderr=add_stderr,
                                 timeout=operation_wait + 10.,
                                 sudo=True)
        except sh.ShellCommandFailed as exc:
            if attempt.is_last:
                raise exc
            else:
                LOG.info('the pcs command failed - retrying...')
                continue
        break
    return output


PCS_COMMAND = sh.shell_command(['pcs'])


def execute_pcs(command_args: list,
                ssh_client: ssh.SSHClientFixture = None,
                pcs_command: sh.ShellCommand = None,
                add_stderr: bool = False,
                timeout: float = 40.,
                **execute_params) -> str:
    if ssh_client is None:
        ssh_client = topology.find_openstack_node(
            group='controller').ssh_client

    if pcs_command:
        pcs_command = sh.shell_command(pcs_command)
    else:
        pcs_command = PCS_COMMAND

    command = pcs_command + command_args
    result = sh.execute(
        command, ssh_client=ssh_client, stdin=False, stdout=True, stderr=True,
        timeout=timeout, **execute_params)

    if add_stderr:
        output = '\n'.join([result.stdout, result.stderr])
    else:
        output = result.stdout
    return output



# Copyright (c) 2023 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from oslo_log import log

import tobiko
from tobiko import config
from tobiko.openstack import octavia
from tobiko.openstack import neutron
from tobiko.openstack.octavia import _constants

CONF = config.CONF
LOG = log.getLogger(__name__)


def get_external_subnet(ip_version=4):
    kw = {'router:external': True}
    if len(neutron.list_networks(**{'router:external': True})) > 1:
        kw['name'] = CONF.tobiko.neutron.external_network

    try:
        ext_subnet_list = neutron.find_network(**kw)['subnets']
    except tobiko.ObjectNotFound:
        LOG.warning('External network not found')
        return None

    for ext_subnet_id in ext_subnet_list:
        try:
            subnet = neutron.find_subnet(id=ext_subnet_id,
                                         ip_version=ip_version)
        except tobiko.ObjectNotFound:
            continue
        else:
            return subnet

    LOG.warning('External subnet with IP version %d not found', ip_version)


def deploy_ipv4_lb(provider: str,
                   protocol: str,
                   protocol_port: int,
                   lb_algorithm: str,
                   servers_stacks=None):
    """Deploy a populated ipv4 LB

    :param provider: the loadbalancer provider. For example: amphora
    :param protocol: the listener & pool protocol. For example: HTTP
    :param protocol_port: the load balancer protocol port
    :param lb_algorithm: the pool load balancing protocol. For example:
        ROUND_ROBIN
    :param servers_stacks: a list of server stacks (until we remove heat
        entirely)
    :return: all Octavia resources it has created (LB, listener, and pool)
    """
    lb_name = octavia.OCTAVIA_PROVIDERS_NAMES['lb'][provider]
    lb = octavia.find_load_balancer(lb_name)
    if lb:
        LOG.debug(f'Loadbalancer {lb.id} already exists. Skipping its'
                  ' creation')
    else:
        subnet = get_external_subnet()
        if subnet is None:
            tobiko.skip_test('Replacing heat networking resources for '
                             'octavia in tobiko wasn\'t implemented yet')
        lb_kwargs = {
            'provider': provider,
            'vip_subnet_id': subnet['id'],
            'name': lb_name
        }
        lb = octavia.create_load_balancer(lb_kwargs)
        octavia.wait_for_status(object_id=lb.id)
        LOG.debug(f'Loadbalancer {lb.name} was deployed successfully '
                  f'with id {lb.id}')

    listener_name = octavia.OCTAVIA_PROVIDERS_NAMES['listener'][provider]
    listener = octavia.find_listener(listener_name)
    if listener:
        LOG.debug(f'Listener {listener.id} already exists. Skipping'
                  ' its creation')
    else:
        listener_kwargs = {
            'protocol': protocol,
            'protocol_port': protocol_port,
            'loadbalancer_id': lb.id,
            'name': listener_name
        }
        listener = octavia.create_listener(listener_kwargs)
        octavia.wait_for_status(object_id=lb.id)
        LOG.debug(f'Listener {listener.name} was deployed '
                  f'successfully with id {listener.id}')

    pool_name = octavia.OCTAVIA_PROVIDERS_NAMES['pool'][provider]
    pool = octavia.find_pool(pool_name)
    if pool:
        LOG.debug(f'Pool {pool.id} already exists. Skipping its '
                  f'creation')
    else:
        pool_kwargs = {
            'listener_id': listener.id,
            'lb_algorithm': lb_algorithm,
            'protocol': protocol,
            'name': pool_name
        }

        pool = octavia.create_pool(pool_kwargs)
        octavia.wait_for_status(object_id=lb.id)
        LOG.debug(f'Pool {pool.name} was deployed successfully with'
                  f' id {pool.id}')

    if servers_stacks:
        for idx, server_stack in enumerate(servers_stacks):
            member_name_prefix = octavia.OCTAVIA_PROVIDERS_NAMES['member'][
                provider]
            member_name = member_name_prefix + str(idx)
            member = octavia.find_member(member_name=member_name,
                                         pool=pool.id)
            if member:
                LOG.debug(f'Member {member.id} already exists. Skipping its '
                          f'creation')
            else:
                member_kwargs = {
                    'address': str(server_stack.fixed_ipv4),
                    'protocol_port': protocol_port,
                    'name': member_name,
                    'subnet_id': server_stack.network_stack.ipv4_subnet_id,
                    'pool': pool.id
                }

                member = octavia.create_member(member_kwargs)
                octavia.wait_for_status(object_id=lb.id)
                LOG.debug(f'Member {member.name} was deployed successfully '
                          f'with id {member.id}')

    return lb, listener, pool


def deploy_ipv4_amphora_lb(protocol: str = _constants.PROTOCOL_HTTP,
                           protocol_port: int = 80,
                           lb_algorithm: str = (
                                   _constants.LB_ALGORITHM_ROUND_ROBIN),
                           servers_stacks=None):
    """Deploy a populated ipv4 amphora provider LB with HTTP resources

    This deployer method deploys the following resources:
        * An IPv4 amphora provider LB
        * An HTTP listener
        * An HTTP pool with Round Robin LB algorithm
        * Octavia members (a member per each nova vm it receives from the
            caller)

    :param protocol: the listener & pool protocol. For example: HTTP
    :param protocol_port: the load balancer protocol port
    :param lb_algorithm: the pool load balancing protocol. For example:
        ROUND_ROBIN
    :param servers_stacks: a list of server stacks (until we remove heat
        entirely)
    :return: all Octavia resources it has created (LB, listener, and pool)
    """
    return deploy_ipv4_lb(provider=octavia.AMPHORA_PROVIDER,
                          protocol=protocol,
                          protocol_port=protocol_port,
                          lb_algorithm=lb_algorithm,
                          servers_stacks=servers_stacks)


def deploy_ipv4_ovn_lb(protocol: str = _constants.PROTOCOL_TCP,
                       protocol_port: int = 80,
                       lb_algorithm: str = (
                               _constants.LB_ALGORITHM_SOURCE_IP_PORT),
                       servers_stacks=None):
    """Deploy a populated ipv4 OVN provider LB with TCP resources

    This deployer method deploys the following resources:
        * An IPv4 ovn provider LB
        * An TCP listener
        * An TCP pool with Source Ip Port LB algorithm
        * Octavia members (a member per each nova vm it receives from the
            caller)

    :param protocol: the listener & pool protocol. For example: TCP
    :param protocol_port: the load balancer protocol port
    :param lb_algorithm: the pool load balancing protocol. For example:
        SOURCE_IP_PORT
    :param servers_stacks: a list of server stacks (until we remove heat
        entirely)
    :return: all Octavia resources it has created (LB, listener, and pool)
    """
    return deploy_ipv4_lb(provider=octavia.OVN_PROVIDER,
                          protocol=protocol,
                          protocol_port=protocol_port,
                          lb_algorithm=lb_algorithm,
                          servers_stacks=servers_stacks)



# Copyright 2020 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from collections import abc
import typing

from oslo_log import log

import tobiko
from tobiko.shell import sh
from tobiko.shell import ssh

LOG = log.getLogger(__name__)


CLOUD_INIT_TRANSIENT_STATES = {
    'done': ('running',)
}

CLOUD_INIT_OUTPUT_FILE = '/var/log/cloud-init-output.log'
CLOUD_INIT_LOG_FILE = '/var/log/cloud-init.log'


def user_data(*args, **kwargs):
    config = cloud_config(*args, **kwargs)
    if config:
        return config.user_data
    else:
        return ''


def cloud_config(*args, **kwargs):
    return combine_cloud_configs(args + (kwargs,))


def combine_cloud_configs(objs):
    packages = []
    runcmd = []
    extra_params = {}
    for obj in objs:
        if obj:
            if not isinstance(obj, abc.Mapping):
                obj = dict(obj)
            for package in obj.pop('packages', []):
                if package and package not in packages:
                    packages.append(package)
            for cmdline in obj.pop('runcmd', []):
                if cmdline:
                    cmdline = list(sh.shell_command(cmdline))
                    if cmdline:
                        runcmd.append(cmdline)
            if obj:
                extra_params.update(obj)

    return CloudConfig.create(packages=packages or None,
                              runcmd=runcmd or None,
                              **extra_params)


class CloudConfig(dict):

    @classmethod
    def create(cls, **kwargs):
        return cls((k, v)
                   for k, v in kwargs.items()
                   if v is not None)

    @property
    def user_data(self):
        return '#cloud-config\n' + tobiko.dump_yaml(dict(self))

    def __add__(self, other):
        return combine_cloud_configs([self, other])

    @property
    def packages(self) -> typing.List[str]:
        return self.setdefault('packages', [])

    def add_package(self, *packages: str):
        _packages = self.packages
        for package in packages:
            tobiko.check_valid_type(package, str)
            if package not in _packages:
                _packages.append(package)

    @property
    def runcmd(self) -> typing.List[typing.List[str]]:
        return self.setdefault('runcmd', [])

    def add_runcmd(self, *command_lines: sh.ShellCommandType):
        runcmd = self.runcmd
        for command_line in command_lines:
            command_line = sh.shell_command(command_line)
            runcmd.append(list(command_line))

    @property
    def write_files(self) -> typing.List[typing.Dict[str, typing.Any]]:
        return self.setdefault('write_files', [])

    def add_write_file(self,
                       path: str,
                       content: str,
                       owner: str = None,
                       permissions: str = None):
        tobiko.check_valid_type(path, str)
        tobiko.check_valid_type(content, str)
        entry = dict(path=path, content=content)
        if owner is not None:
            tobiko.check_valid_type(owner, str)
            entry['owner'] = owner
        if permissions is not None:
            tobiko.check_valid_type(owner, str)
            entry['permission'] = permissions
        self.write_files.append(entry)


class InvalidCloudInitStatusError(tobiko.TobikoException):
    message = ("cloud-init status of host '{hostname}' is "
               "'{actual_status}' while it is expecting to "
               "be in {expected_states!r}:\n\n"
               f"--- {CLOUD_INIT_LOG_FILE} ---\n"
               "{log_file}\n\n"
               f"--- {CLOUD_INIT_OUTPUT_FILE} ---\n"
               "{output_file}\n\n")


class WaitForCloudInitTimeoutError(InvalidCloudInitStatusError):
    message = ("after {timeout} seconds cloud-init status of host "
               "'{hostname}' is still '{actual_status}' while it is "
               "expecting to be in {expected_states!r}:\n\n"
               f"--- {CLOUD_INIT_LOG_FILE} ---\n"
               "{log_file}\n\n"
               f"--- {CLOUD_INIT_OUTPUT_FILE} ---\n"
               "{output_file}\n\n")


def get_cloud_init_status(
        ssh_client: typing.Optional[ssh.SSHClientFixture] = None,
        timeout: tobiko.Seconds = None) \
        -> str:
    try:
        output = sh.execute('cloud-init status',
                            ssh_client=ssh_client,
                            timeout=timeout,
                            sudo=True).stdout
    except sh.ShellCommandFailed as ex:
        output = ex.stdout
        if output:
            LOG.debug(f"Cloud init status error reported:\n{ex}")
        else:
            raise

    status = tobiko.load_yaml(output)
    tobiko.check_valid_type(status, dict)
    tobiko.check_valid_type(status['status'], str)
    return status['status']


def wait_for_cloud_init_done(
        ssh_client: typing.Optional[ssh.SSHClientFixture] = None,
        timeout: tobiko.Seconds = None,
        sleep_interval: tobiko.Seconds = None) \
        -> str:
    return wait_for_cloud_init_status('done',
                                      ssh_client=ssh_client,
                                      timeout=timeout,
                                      sleep_interval=sleep_interval)


def wait_for_cloud_init_status(
        *expected_states: str,
        transient_states: typing.Optional[typing.Container[str]] = None,
        ssh_client: typing.Optional[ssh.SSHClientFixture] = None,
        timeout: tobiko.Seconds = None,
        sleep_interval: tobiko.Seconds = None) \
        -> str:
    hostname = sh.get_hostname(ssh_client=ssh_client,
                               timeout=timeout)
    if transient_states is None:
        transient_states = list()
        for status in expected_states:
            transient_states += CLOUD_INIT_TRANSIENT_STATES.get(status, [])

    def _read_file(filename: str, tail=False) -> str:
        return read_file(filename=filename,
                         ssh_client=ssh_client,
                         timeout=timeout,
                         tail=tail)

    actual_status: typing.Optional[str]

    for attempt in tobiko.retry(timeout=timeout,
                                interval=sleep_interval,
                                default_timeout=1200.,
                                default_interval=5.):
        try:
            actual_status = get_cloud_init_status(ssh_client=ssh_client,
                                                  timeout=attempt.time_left)
        except sh.ShellCommandFailed:
            LOG.exception('Unable to get cloud-init status')
            actual_status = None
        else:
            if actual_status in expected_states:
                break

        if attempt.is_last:
            raise WaitForCloudInitTimeoutError(
                timeout=attempt.timeout,
                hostname=hostname,
                actual_status=actual_status,
                expected_states=expected_states,
                log_file=_read_file(CLOUD_INIT_LOG_FILE),
                output_file=_read_file(CLOUD_INIT_OUTPUT_FILE))

        elif actual_status in transient_states:
            last_log_lines = _read_file(CLOUD_INIT_LOG_FILE, tail=True)
            LOG.debug(f"Waiting cloud-init status on host '{hostname}' to "
                      f"switch from '{actual_status}' to any of expected "
                      f"states ({', '.join(expected_states)}):\n\n"
                      f"--- {CLOUD_INIT_LOG_FILE} ---\n"
                      f"{last_log_lines}\n\n")
        else:
            raise InvalidCloudInitStatusError(
                hostname=hostname,
                actual_status=actual_status,
                expected_states=expected_states,
                log_file=_read_file(CLOUD_INIT_LOG_FILE),
                output_file=_read_file(CLOUD_INIT_OUTPUT_FILE))

    else:
        raise RuntimeError('Broken retry loop')
    return actual_status


def read_file(filename: str,
              tail=False,
              ssh_client: ssh.SSHClientType = None,
              timeout: tobiko.Seconds = None):
    if tail:
        command = 'tail'
    else:
        command = 'cat'
    try:
        return sh.execute(f'{command} "{filename}"',
                          timeout=timeout,
                          ssh_client=ssh_client).stdout
    except sh.ShellCommandFailed:
        LOG.exception(f"Error reading file '{filename}'")
        return ""



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os
import random
import string
import typing

import testtools

import tobiko
from tobiko.openstack import heat
from tobiko.openstack import keystone
from tobiko.openstack import neutron
from tobiko.openstack import nova
from tobiko.openstack import stacks


TEMPLATE_DIRS = [os.path.dirname(__file__)]


def random_string(length, letters=string.ascii_lowercase):
    return ''.join(random.choice(letters) for _ in range(length))


class MyStack(heat.HeatStackFixture):

    template = heat.heat_template_file(template_file='hot/my_stack.yaml',
                                       template_dirs=TEMPLATE_DIRS)

    input_text = random_string(8)


class HeatStackFixtureTest(testtools.TestCase):

    stack = tobiko.required_fixture(MyStack)

    def test_get_stack(self):
        self.stack.wait_for_create_complete()
        stack = self.stack.get_stack()
        self.assertIsNotNone(stack)
        self.assertEqual(tobiko.get_fixture_name(MyStack), stack.stack_name)
        self.assertIsInstance(stack.id, str)
        self.assertIsInstance(stack.stack_status, str)

    def test_get_fixture_with_fixture_id_0(self):
        fixture_0 = tobiko.get_fixture(MyStack, fixture_id=0)
        self.assertIs(fixture_0, self.stack)

    def test_get_fixture_with_fixture_id_1(self):
        fixture_0 = tobiko.get_fixture(MyStack)
        fixture_1 = tobiko.get_fixture(MyStack, fixture_id=1)
        self.assertIsNot(fixture_0, fixture_1)
        stack_0 = tobiko.setup_fixture(fixture_0).get_stack()
        stack_1 = tobiko.setup_fixture(fixture_1).get_stack()
        self.assertNotEqual(stack_0.id, stack_1.id)
        self.assertEqual(tobiko.get_fixture_name(MyStack), stack_0.stack_name)
        self.assertEqual(tobiko.get_fixture_name(MyStack) + '-1',
                         stack_1.stack_name)


class EnsureNeutronQuotaLimitsFixture(MyStack):

    requirements = {'network': 100, 'subnet': 100, 'router': 10}

    @property
    def neutron_required_quota_set(self) -> typing.Dict[str, int]:
        return self.requirements


class EnsureNovaQuotaLimitsFixture(MyStack):

    requirements = {'cores': 20, 'instances': 10}

    @property
    def nova_required_quota_set(self) -> typing.Dict[str, int]:
        return self.requirements


class EnsureQuotaLimitsTest(testtools.TestCase):

    def test_ensure_neutron_quota_limits(self):
        stack = EnsureNeutronQuotaLimitsFixture(stack_name=self.id())
        self.useFixture(stack)
        quota_set = neutron.get_neutron_quota_set(detail=True)
        for name, requirement in stack.requirements.items():
            quota = quota_set[name]
            self.assertGreaterEqual(int(quota['limit']),
                                    requirement +
                                    max(0, int(quota['used'])) -
                                    max(0, int(quota['reserved'])))

    def test_ensure_nova_quota_limits(self):
        stack = EnsureNovaQuotaLimitsFixture(stack_name=self.id())
        self.useFixture(stack)
        quota_set = nova.get_nova_quota_set(detail=True)
        for name, requirement in stack.requirements.items():
            quota = getattr(quota_set, name)
            if int(quota['limit']) > 0:
                self.assertGreaterEqual(int(quota['limit']),
                                        int(requirement) +
                                        max(0, int(quota['in_use'])) -
                                        max(0, int(quota['reserved'])))
            else:
                self.assertEqual(int(quota['limit']), -1)


@keystone.skip_unless_has_keystone_credentials()
class StackTest(testtools.TestCase):

    router_stack = tobiko.required_fixture(stacks.RouterStackFixture)

    def test_find_stack(self):
        stack = self.router_stack.stack
        self.assertIsInstance(stack, heat.STACK_CLASSES)
        result = heat.find_stack(name=stack.stack_name)
        self.assertEqual(stack.stack_name, result.stack_name)

    def test_list_stacks(self):
        stack = self.router_stack.stack
        result = heat.list_stacks()
        self.assertIsInstance(result, tobiko.Selection)
        self.assertEqual(stack.id,
                         result.with_attributes(id=stack.id).unique.id)


@keystone.skip_unless_has_keystone_credentials()
class ResourceTest(testtools.TestCase):

    router_stack = tobiko.required_fixture(stacks.RouterStackFixture)

    def test_find_resource(self):
        self.router_stack.wait_for_create_complete()
        resource = heat.list_resources(stack=self.router_stack).first
        result = heat.find_resource(
            stack=self.router_stack,
            physical_resource_id=resource.physical_resource_id)
        self.assertIsInstance(resource, heat.RESOURCE_CLASSES)
        self.assertEqual(resource.physical_resource_id,
                         result.physical_resource_id)

    def test_list_resource(self):
        self.router_stack.wait_for_create_complete()
        resources = heat.list_resources(stack=self.router_stack)
        self.assertIsInstance(resources, tobiko.Selection)
        self.assertNotEqual([], resources)
        self.assertEqual(resources,
                         resources.with_attributes(
                             stack_name=self.router_stack.stack_name))



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import io
import os
import time
import typing  # noqa
from multiprocessing import Process as MultiProcess

import psutil
from oslo_log import log


import tobiko
from tobiko.shell import sh
from tobiko.shell.sh import _command
from tobiko.shell.sh import _exception
from tobiko.shell.sh import _io


LOG = log.getLogger(__name__)


def process(command=None, environment=None, timeout: tobiko.Seconds = None,
            shell=None, stdin=None, stdout=None, stderr=None, ssh_client=None,
            sudo=None, **kwargs):
    kwargs.update(command=command, environment=environment, timeout=timeout,
                  shell=shell, stdin=stdin, stdout=stdout, stderr=stderr,
                  sudo=sudo)
    if timeout is not None:
        if timeout < 0.:
            raise ValueError("Invalid timeout for executing process: "
                             "{!r}".format(timeout))
    try:
        from tobiko.shell.sh import _ssh
        from tobiko.shell import ssh
    except ImportError:
        if ssh_client:
            raise
    else:
        if ssh_client is None:
            ssh_client = ssh.ssh_proxy_client()
        if ssh_client:
            return _ssh.ssh_process(ssh_client=ssh_client, **kwargs)

    from tobiko.shell.sh import _local
    return _local.local_process(**kwargs)


class Parameters(object):

    def __init__(self, **kwargs):
        cls = type(self)
        for name, value in kwargs.items():
            if value is not None:
                if not hasattr(cls, name):
                    raise ValueError('Invalid parameter: {!s}'.format(name))
                setattr(self, name, value)


class ShellProcessParameters(Parameters):

    command = None
    environment = None
    current_dir = None
    timeout: tobiko.Seconds = None
    stdin = False
    stdout = True
    stderr = True
    buffer_size = io.DEFAULT_BUFFER_SIZE
    poll_interval = 1.
    network_namespace = None
    retry_count: typing.Optional[int] = 3
    retry_interval: tobiko.Seconds = 5.
    retry_timeout: tobiko.Seconds = 120.
    shell: typing.Union[None, bool, str] = None
    sudo: typing.Union[None, bool, str] = None


class ShellProcessFixture(tobiko.SharedFixture):

    command = None
    process: typing.Any = None
    stdin = None
    stdout = None
    stderr = None
    default_shell: typing.Union[None, bool, str] = None

    _exit_status = None

    def __init__(self, **kwargs):
        super(ShellProcessFixture, self).__init__()
        self.parameters = self.init_parameters(**kwargs)

    def init_parameters(self, **kwargs) -> ShellProcessParameters:
        return ShellProcessParameters(**kwargs)

    def execute(self) -> 'ShellProcessFixture':
        return tobiko.setup_fixture(self)

    def setup_fixture(self):
        parameters = self.parameters

        self.setup_command()
        self.setup_process()

        if parameters.stdin:
            self.setup_stdin()
        if parameters.stdout:
            self.setup_stdout()
        if parameters.stderr:
            self.setup_stderr()

    def setup_command(self):
        command = _command.shell_command(self.parameters.command)
        network_namespace = self.parameters.network_namespace
        sudo = self.parameters.sudo
        shell = self.parameters.shell
        if shell is None:
            shell = self.default_shell

        if shell is not None:
            tobiko.check_valid_type(shell, (bool, str))
            if isinstance(shell, str):
                command = _command.shell_command(shell) + [str(command)]
            elif shell is True:
                command = default_shell_command() + [str(command)]

        if network_namespace:
            if sudo is None:
                sudo = True
            command = network_namespace_command(network_namespace, command)

        if sudo:
            if sudo is True:
                sudo = default_sudo_command()
            else:
                sudo = _command.shell_command(sudo)
            command = sudo + command

        self.command = command

    def setup_process(self):
        if self._exit_status:
            del self._exit_status
        self.process = self.create_process()
        self.addCleanup(self.close)

    def setup_stdin(self):
        raise NotImplementedError

    def setup_stdout(self):
        raise NotImplementedError

    def setup_stderr(self):
        raise NotImplementedError

    def create_process(self):
        raise NotImplementedError

    def close_stdin(self):
        stdin = self.stdin
        if stdin is not None:
            try:
                stdin.closed or stdin.close()
            except Exception:
                LOG.exception("Error closing STDIN stream: %r", self.stdin)

    def close_stdout(self):
        stdout = self.stdout
        if stdout is not None:
            try:
                stdout.closed or stdout.close()
            except Exception:
                LOG.exception("Error closing STDOUT stream: %r", self.stdout)

    def close_stderr(self):
        stderr = self.stderr
        if stderr is not None:
            try:
                stderr.closed or stderr.close()
            except Exception:
                LOG.exception("Error closing STDERR stream: %r", self.stderr)

    def close(self, timeout: tobiko.Seconds = None):
        if timeout is None:
            timeout = self.parameters.timeout
        self.close_stdin()
        try:
            # Drain all incoming data from STDOUT and STDERR
            self.wait(timeout=timeout)
        finally:
            self._terminate()

    def _terminate(self):
        self.close_stdout()
        self.close_stderr()
        exit_status = None
        try:
            exit_status = self.get_exit_status()
        finally:
            if exit_status is None:
                try:
                    self.kill()
                except Exception:
                    LOG.exception('Error killing process: %r', self.command)

    def __getattr__(self, name):
        try:
            # Get attributes from parameters class
            return getattr(self.parameters, name)
        except AttributeError as ex:
            message = "object {!r} has not attribute {!r}".format(self, name)
            raise AttributeError(message) from ex

    def kill(self, sudo=False):
        raise NotImplementedError

    def poll_exit_status(self):
        raise NotImplementedError

    def get_exit_status(self, timeout: tobiko.Seconds = None):
        if timeout is None:
            timeout = self.parameters.timeout
        exit_status = self._get_exit_status(timeout=timeout)
        if exit_status is not None:
            return exit_status

        ex = _exception.ShellTimeoutExpired(
            command=str(self.command),
            timeout=timeout,
            stdin=str_from_stream(self.stdin),
            stdout=str_from_stream(self.stdout),
            stderr=str_from_stream(self.stderr))
        LOG.debug("Timed out while waiting for command termination:\n%s",
                  self.command)
        raise ex

    def _get_exit_status(self, timeout):
        raise NotImplementedError

    @property
    def exit_status(self):
        exit_status = self._exit_status
        if exit_status is None:
            exit_status = self.poll_exit_status()
            if exit_status is not None:
                self._exit_status = exit_status
        return exit_status

    @property
    def is_running(self):
        return self.exit_status is None

    def check_is_running(self):
        exit_status = self.exit_status
        if exit_status is not None:
            raise _exception.ShellProcessTerminated(
                command=str(self.command),
                exit_status=int(exit_status),
                stdin=str_from_stream(self.stdin),
                stdout=str_from_stream(self.stdout),
                stderr=str_from_stream(self.stderr))

    def check_stdin_is_opened(self):
        if self.stdin.closed:
            raise _exception.ShellStdinClosed(
                command=str(self.command),
                stdin=str_from_stream(self.stdin),
                stdout=str_from_stream(self.stdout),
                stderr=str_from_stream(self.stderr))

    def send_all(self, data, **kwargs):
        self.communicate(stdin=data, **kwargs)
        self.stdin.flush()

    def receive_all(self, **kwargs):
        self.communicate(receive_all=True, **kwargs)

    def wait(self, timeout: tobiko.Seconds = None, receive_all=True,
             **kwargs):
        self.communicate(timeout=timeout, receive_all=receive_all,
                         **kwargs)

    def communicate(self, stdin=None, stdout=True, stderr=True,
                    timeout: tobiko.Seconds = None,
                    receive_all=False, buffer_size=None):
        timeout = tobiko.to_seconds(timeout)

        # Avoid waiting for data in the first loop
        poll_interval = 0.
        streams = _io.select_opened_files([stdin and self.stdin,
                                           stdout and self.stdout,
                                           stderr and self.stderr])
        for attempt in tobiko.retry(timeout=timeout):
            if not self._is_communicating(streams=streams, send=stdin,
                                          receive=receive_all):
                break

            # Remove closed streams
            streams = _io.select_opened_files(streams)

            # Select ready streams
            read_ready, write_ready = _io.select_files(
                files=streams, timeout=poll_interval)
            if read_ready or write_ready:
                # Avoid waiting for data the next time
                poll_interval = 0.
                if self.stdin in write_ready:
                    # Write data to remote STDIN
                    stdin = self._write_to_stdin(stdin)
                    if not stdin:
                        streams.remove(self.stdin)
                if self.stdout in read_ready:
                    # Read data from remote STDOUT
                    stdout = self._read_from_stdout(buffer_size=buffer_size)
                    if not stdout:
                        streams.remove(self.stdout)
                if self.stderr in read_ready:
                    # Read data from remote STDERR
                    stderr = self._read_from_stderr(buffer_size=buffer_size)
                    if not stderr:
                        streams.remove(self.stderr)
            else:
                self._check_communicate_timeout(attempt=attempt,
                                                timeout=timeout)
                # Wait for data in the following loops
                poll_interval = self.parameters.poll_interval
                LOG.debug(f"Waiting for process data {poll_interval} "
                          f"seconds... \n"
                          f"  command: {self.command}\n"
                          f"  attempt: {attempt.details}\n"
                          f"  streams: {streams}")

    def _check_communicate_timeout(self, attempt: tobiko.RetryAttempt,
                                   timeout: tobiko.Seconds):
        try:
            attempt.check_limits()
        except tobiko.RetryTimeLimitError:
            LOG.exception("retry timeout expired")
            # Eventually raises either ShellCommandTimeout exception or
            # RetryTimeLimitError
            self.get_exit_status(timeout=timeout)
            raise
        else:
            return

    def _is_communicating(self, streams, send, receive):
        if send and self.stdin in streams:
            return True
        elif receive and {self.stdout, self.stderr} & streams:
            return True
        else:
            return False

    def _write_to_stdin(self, data, check=True):
        """Write data to STDIN"""
        if check:
            self.check_stdin_is_opened()
        sent_bytes = self.stdin.write(data)
        if sent_bytes:
            return data[sent_bytes:] or None
        else:
            LOG.debug("%r closed by peer on %r", self.stdin, self)
            self.stdin.close()
            return data

    def _read_from_stdout(self, buffer_size=None):
        """Read data from remote stream"""
        # Read data from remote stream
        chunk = self.stdout.read(buffer_size)
        if chunk:
            return chunk
        else:
            LOG.debug("%r closed by peer on %r", self.stdout, self)
            self.stdout.close()
            return None

    def _read_from_stderr(self, buffer_size=None):
        """Read data from remote stream"""
        # Read data from remote stream
        chunk = self.stderr.read(buffer_size)
        if chunk:
            return chunk
        else:
            LOG.debug("%r closed by peer on %r", self.stderr, self)
            self.stderr.close()
            return None

    def check_exit_status(self, expected_status=0):
        exit_status = self.poll_exit_status()
        if exit_status is None:
            time_left = self.check_timeout()
            ex = _exception.ShellProcessNotTerminated(
                command=str(self.command),
                time_left=time_left,
                stdin=self.stdin,
                stdout=self.stdout,
                stderr=self.stderr)
            raise ex

        exit_status = int(exit_status)
        if expected_status != exit_status:
            ex = _exception.ShellCommandFailed(
                command=str(self.command),
                exit_status=exit_status,
                stdin=str_from_stream(self.stdin),
                stdout=str_from_stream(self.stdout),
                stderr=str_from_stream(self.stderr))
            raise ex


def merge_dictionaries(*dictionaries):
    merged = {}
    for d in dictionaries:
        if d:
            merged.update(d)
    return merged


def str_from_stream(stream):
    if stream is not None:
        try:
            return str(stream)
        except UnicodeDecodeError:
            LOG.exception('Unable to decode as a string - '
                          'Returning the raw data')
            return stream.data
    else:
        return None


def bytes_from_stream(stream):
    if stream is not None:
        return stream.data
    else:
        return None


def default_shell_command():
    from tobiko import config
    CONF = config.CONF
    return _command.shell_command(CONF.tobiko.shell.command)


def default_sudo_command():
    from tobiko import config
    CONF = config.CONF
    return _command.shell_command(CONF.tobiko.shell.sudo)


def network_namespace_command(network_namespace, command):
    return _command.shell_command(['/sbin/ip', 'netns', 'exec',
                                   network_namespace]) + command


def start_background_process(bg_function=None, bg_process_name=None, **kwargs):
    """Background process that will take a function name as parameter
    and execute it in the background using a separate non attached process.
    That process will continue to run even after Tobiko exists.
    params:
    bg_function= function name to run in background
    bg_process_pid_file= file path that will contain the process pid, multiple
    processes can use the same file pid are appended.
    outputs: writes processes pids to a file, each in a line
    returns: the process object"""

    # define a parent process that would be killed and orphan the actual
    # background process to run unattached in the background
    # this is so the background process won't be stopped when tobiko exists
    def _background_process_parent():
        p = MultiProcess(target=bg_function, name=bg_process_name,
                         kwargs=kwargs)
        p.start()
        LOG.info(
            f'Started background function: {bg_function.__name__} process pid '
            f'is: {p.pid}, process name: {bg_process_name}, '
            f'main execution process continues...')
        # append bg_process pid to a file
        bg_process_pids_file_name = f'{sh.get_user_home_dir()}/' \
                                    f'{bg_process_name}_pids_file'
        with open(bg_process_pids_file_name, "at") as bg_process_pid_file:
            bg_process_pid_file.write(str(p.pid) + "\n")
            LOG.debug(f'Writing pid: {p.pid} to pids file:'
                      f' {bg_process_pids_file_name}')

    # start parent process, nested with a started child process
    # then kill the parent
    d = MultiProcess(target=_background_process_parent)
    d.daemon = False
    d.start()
    LOG.debug(f'Background process parent started pid: {d.pid}')
    time.sleep(1)
    d.terminate()
    LOG.debug(f'Background process orphaned,  parent killed parent pid:'
              f' {d.pid}')


def stop_process(pid_list):
    """Stop (kill) a process from a list"""
    for pid in pid_list:

        LOG.info(f'stopping process with pid: {pid}')
        sh.execute(f'sudo kill -9 {pid}')


def get_bg_procs_pids(bg_process_name):
    """return a list of pids from the specified bg_process_name file"""
    bg_process_pids_file_name = f'{sh.get_user_home_dir()}/' \
                                f'{bg_process_name}_pids_file'
    bg_process_name_pid_list = []
    if os.path.isfile(bg_process_pids_file_name):
        LOG.info(f'found previous background process file :'
                 f' {bg_process_pids_file_name}, cheking it`s processes.')
        # go over file's pids
        with io.open(bg_process_pids_file_name, 'rt') as fd:
            for line in fd.readlines():
                pid = line.rstrip()
                try:
                    proc = psutil.Process(int(pid))
                # continue if pid is not a valid int or doesn't exist
                except (TypeError, ValueError, psutil.NoSuchProcess):
                    continue
                # check if process is running
                if proc.status() != psutil.STATUS_ZOMBIE:
                    LOG.debug(f'skipping process {pid} , it\'s a zombie')
                    bg_process_name_pid_list.append(pid)
    return bg_process_name_pid_list


def check_or_start_background_process(bg_function=None,
                                      bg_process_name=None,
                                      check_function=None,
                                      **kwargs):
    """Start or restart a process.

    Check if a process exists, if so restart the process, execute some check
    logic i.e. a check function. if the process by name isn't running, start a
    new separate process i.e a background function

    :param bg_process_name: process name
    :param bg_function: function name
    :param check_function: function name
    """

    procs_running_list = get_bg_procs_pids(bg_process_name)
    if procs_running_list:
        # in any case test is still running, check for failures:
        # execute process check i.e. go over process results file
        # truncate the log file and restart the background process
        LOG.info(f'running a check function: {check_function} '
                 f'on results of processes: {bg_process_name}')
        check_function()
        # if we want to terminate the specific background process by
        # name, close it, otherwise the check will continue to run in the
        # background
        stop_process(procs_running_list)
        LOG.info('checked and stopped previous background processes and log '
                 'starting a new background process ')
    else:
        # First time the test is run:
        # if background process by specific name is not present ,
        # start one in the background:
        LOG.info(f'No previous background processes found:'
                 f' {bg_process_name}, starting a new background process '
                 f'of function: {bg_function}')

    start_background_process(bg_function=bg_function,
                             bg_process_name=bg_process_name, **kwargs)
    # check test is not failing from the start
    check_function()



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import netaddr
from oslo_log import log

import tobiko
from tobiko.shell import sh


LOG = log.getLogger(__name__)


class IpError(tobiko.TobikoException):
    message = ("Unable to get IP addresses from host "
               "(exit_status={exit_status!r}): {error!s}")


INETS = {
    4: ['inet'],
    6: ['inet6'],
    None: ['inet', 'inet6']
}


def list_ip_addresses(ip_version: int = None,
                      device: str = None,
                      scope: str = None,
                      **execute_params) -> \
        tobiko.Selection[netaddr.IPAddress]:
    inets = INETS.get(ip_version)
    if inets is None:
        error = "invalid IP version: {!r}".format(ip_version)
        raise IpError(error=error)

    command = ['-o', 'address', 'list']
    if device is not None:
        tobiko.check_valid_type(device, str)
        command.append(device)
    output = execute_ip(command, **execute_params)

    ips: tobiko.Selection[netaddr.IPAddress] = tobiko.Selection()
    if output:
        for line in output.splitlines():
            fields = line.strip().split()
            inet = fields[2]
            if inet not in inets:
                continue  # List only address of selected IP version

            if scope:
                try:
                    scope_index = fields.index('scope')
                    if fields[scope_index + 1] != scope:
                        continue
                except (IndexError, ValueError):
                    continue

            address, _ = parse_ip_address(fields[3])
            ips.append(address)
    return ips


def find_ip_address(ip_version: int = None,
                    device: str = None,
                    scope: str = None,
                    unique: bool = False,
                    **execute_params) -> \
        netaddr.IPAddress:
    ips = list_ip_addresses(ip_version=ip_version,
                            device=device,
                            scope=scope,
                            **execute_params)
    if unique:
        return ips.unique
    else:
        return ips.first


def parse_ip_address(text: str) -> typing.Tuple[netaddr.IPAddress, int]:
    address = text
    if '/' in text:
        # Remove netmask prefix length
        address, prefix_len_text = text.split('/', 1)
        prefix_len = int(prefix_len_text)
    else:
        prefix_len = 0
    return netaddr.IPAddress(address), prefix_len


def list_network_namespaces(**execute_params) -> tobiko.Selection[str]:
    namespaces = tobiko.Selection[str]()
    output = execute_ip(['-o', 'netns', 'list'], **execute_params)
    if output:
        for line in output.splitlines():
            fields = line.strip().split()
            namespace = fields[0]
            namespaces.append(namespace)
    return namespaces


def list_network_interfaces(**execute_params) -> tobiko.Selection[str]:
    interfaces = tobiko.Selection[str]()
    output = execute_ip(['--brief', 'address', 'list'], **execute_params)
    if output:
        for line in output.splitlines():
            fields = line.strip().split()
            interface = fields[0]
            interfaces.append(interface)
    return interfaces


def get_network_main_route_device(dest_ip, **execute_params):
    output = execute_ip(['route', 'get', dest_ip], **execute_params)
    if output:
        for line in output.splitlines():
            fields = line.strip().split()
            device_index = fields.index('dev') + 1
            return fields[device_index]


IP_COMMAND = sh.shell_command(['/sbin/ip'])


def execute_ip(ifconfig_args, ip_command=None, ignore_errors=False,
               **execute_params) -> str:
    if ip_command:
        ip_command = sh.shell_command(ip_command)
    else:
        ip_command = IP_COMMAND
    command = ip_command + ifconfig_args
    result = sh.execute(command, stdin=False, stdout=True, stderr=True,
                        expect_exit_status=None, **execute_params)
    if not ignore_errors and result.exit_status:
        raise IpError(error=result.stderr, exit_status=result.exit_status)
    return result.stdout



"""Models for accessing details from varlink server."""

from __future__ import absolute_import

import collections

import pkg_resources

from . import cached_property


class System():
    """Model for accessing system resources."""

    def __init__(self, client):
        """Construct system model."""
        self._client = client

    @cached_property
    def versions(self):
        """Access versions."""
        with self._client() as podman:
            vers = podman.GetVersion()

        client = '0.0.0'
        try:
            client = pkg_resources.get_distribution('podman').version
        except Exception:  # pylint: disable=broad-except
            pass
        vers['client_version'] = client
        return collections.namedtuple('Version', vers.keys())(**vers)

    def info(self):
        """Return podman info."""
        with self._client() as podman:
            info = podman.GetInfo()['info']
        return collections.namedtuple('Info', info.keys())(**info)

    def ping(self):
        """Return True if server awake."""
        with self._client() as podman:
            response = podman.GetVersion()
        return 'version' in response



"""A client for communicating with a Podman server."""

from __future__ import absolute_import

from pbr.version import VersionInfo

from .client import Client
from .libs import FoldedString, datetime_format, datetime_parse
from .libs.errors import (ContainerNotFound, ErrorOccurred, ImageNotFound,
                          InvalidState, NoContainerRunning, NoContainersInPod,
                          PodContainerError, PodmanError, PodNotFound)

assert FoldedString

try:
    __version__ = VersionInfo("podman")
except Exception:  # pylint: disable=broad-except
    __version__ = '0.0.0'

__all__ = [
    'Client',
    'ContainerNotFound',
    'datetime_format',
    'datetime_parse',
    'ErrorOccurred',
    'ImageNotFound',
    'InvalidState',
    'NoContainerRunning',
    'NoContainersInPod',
    'PodContainerError',
    'PodmanError',
    'PodNotFound',
]



# Copyright (c) 2023 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import neutron
from tobiko.openstack.stacks._neutron import SubnetPoolFixture

PREFIX = "10.211.{i}.0/24"
DEFAULT_PREFIXLEN = 29


@keystone.skip_unless_has_keystone_credentials()
class SubnetPoolTest(testtools.TestCase):
    """Tests subnet pool creation"""

    stack = tobiko.required_fixture(SubnetPoolFixture,
                                    fixture_id="thistest",
                                    prefixes=[PREFIX.format(i=0)],
                                    default_prefixlen=DEFAULT_PREFIXLEN)

    @classmethod
    def tearDownClass(cls):
        # NOTE: the skip at class level does not affect tearDownClass, so
        # the following workaround is needed to avoid errors
        if keystone.has_keystone_credentials():
            tobiko.cleanup_fixture(cls.stack.fixture)

    def test_subnet_pool_find(self):
        snp = neutron.list_subnet_pools(name=self.stack.name).unique
        self.assertEqual(snp, self.stack.subnet_pool)

    def test_subnet_pool_parameters(self):
        self.assertEqual([PREFIX.format(i=0)],
                         self.stack.subnet_pool['prefixes'])
        self.assertEqual(str(DEFAULT_PREFIXLEN),
                         self.stack.subnet_pool['default_prefixlen'])


NUM_SUBNET_POOLS = 10


@keystone.skip_unless_has_keystone_credentials()
class SubnetPoolListTest(testtools.TestCase):
    """Tests creation of a list of subnet pools"""

    stack_list = [tobiko.required_fixture(SubnetPoolFixture,
                                          fixture_id=str(i),
                                          prefixes=[PREFIX.format(i=i)],
                                          default_prefixlen=DEFAULT_PREFIXLEN)
                  for i in range(NUM_SUBNET_POOLS)]
    snp_fixture_list: list = []

    @classmethod
    def tearDownClass(cls):
        for stack in cls.stack_list:
            tobiko.cleanup_fixture(stack.fixture)

    @classmethod
    def setUpClass(cls):
        for stack in cls.stack_list:
            cls.snp_fixture_list.append(tobiko.use_fixture(stack.fixture))

    def test_subnet_pool_list_find(self):
        self.assertEqual(NUM_SUBNET_POOLS, len(self.snp_fixture_list))
        for i, snp_fixture in enumerate(self.snp_fixture_list):
            snp_name = (f"{SubnetPoolFixture.__module__}."
                        f"{SubnetPoolFixture.__qualname__}-{i}")
            self.assertEqual(snp_name, snp_fixture.name)
            snp = neutron.list_subnet_pools(name=snp_name).unique
            self.assertEqual(snp, snp_fixture.subnet_pool)

    def test_subnet_pool_list_parameters(self):
        for i, snp_fixture in enumerate(self.snp_fixture_list):
            self.assertEqual([PREFIX.format(i=i)],
                             snp_fixture.subnet_pool['prefixes'])
            self.assertEqual(str(DEFAULT_PREFIXLEN),
                             snp_fixture.subnet_pool['default_prefixlen'])



#!/usr/bin/env python3
# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os
import sys

TOP_DIR = os.path.realpath(os.path.dirname(os.path.dirname(__file__)))

if TOP_DIR not in sys.path:
    sys.path.insert(0, TOP_DIR)

from tools import common  # noqa


def main():
    version = get_version()
    sys.stdout.write(f'{version}\n')


CACHE = {}


def get_version():
    version = CACHE.get('version')
    if not version:
        CACHE['version'] = version = common.execute(
            f"git -C '{TOP_DIR}' describe --always").splitlines()[0]
    return version


if __name__ == '__main__':
    main()



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.shiftstack import _clouds_file
from tobiko.shiftstack import _neutron
from tobiko.shiftstack import _nova
from tobiko.shiftstack import _skip

ShiftStackCloudsFileFixture = _clouds_file.ShiftStackCloudsFileFixture
get_clouds_file_path = _clouds_file.get_clouds_file_path

list_shiftstack_node_ip_addresses = _neutron.list_shiftstack_node_ip_addresses
find_shiftstack_node_ip_address = _neutron.find_shiftstack_node_ip_address

find_shiftstack_node = _nova.find_shiftstack_node
list_shiftstack_nodes = _nova.list_shiftstack_nodes

skip_unless_has_shiftstack = _skip.skip_unless_has_shiftstack



# Copyright (c) 2022 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

from oslo_log import log

import tobiko
from tobiko import config
from tobiko.openstack import heat
from tobiko.openstack import designate
from tobiko.openstack.stacks import _hot

CONF = config.CONF
LOG = log.getLogger(__name__)


class DesignateZoneStackFixture(heat.HeatStackFixture):
    template = _hot.heat_template_file('designate/zone.yaml')

    record_name = 'record'

    record_type = 'A'

    recordset_records = ['192.10.1.1']

    @property
    def zone_name(self) -> str:
        return tobiko.get_fixture_name(self).lower() + '.'

    @property
    def zone_details(self) -> typing.Mapping[str, typing.Any]:
        return designate.get_designate_zone(self.zone_id)

    @property
    def recordset_list(self) -> typing.Mapping[str, typing.Any]:
        return designate.list_recordsets(self.zone_id)

    def recordset_create(self):
        return designate.create_recordsets(
            self.zone_id,
            name=self.record_name,
            type_=self.record_type,
            records=self.recordset_records)

    def wait_for_active_recordsets(self):
        for recordset in designate.list_recordsets(self.zone_id):
            self.wait_for_active_recordset(zone_id=self.zone_id,
                                           recordset_id=recordset['id'])

    def wait_for_active_recordset(self, zone_id, recordset_id, **kwargs):
        designate.wait_for_status(
            status_key=designate.STATUS,
            status=designate.ACTIVE,
            get_client=designate.get_recordset,
            object_id=zone_id,
            recordset_id=recordset_id, **kwargs
        )



# Copyright 2023 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from collections import abc
import typing

import tobiko
from tobiko.openstack.neutron import _client


SubnetPoolType = typing.Dict[str, typing.Any]
SubnetPoolIdType = typing.Union[str, SubnetPoolType]


def get_subnet_pool_id(subnet_pool: SubnetPoolIdType) -> str:
    if isinstance(subnet_pool, str):
        return subnet_pool
    else:
        return subnet_pool['id']


def get_subnet_pool(subnet_pool: SubnetPoolIdType,
                    client: _client.NeutronClientType = None,
                    **params) -> SubnetPoolType:
    subnet_pool_id = get_subnet_pool_id(subnet_pool)
    try:
        return _client.neutron_client(client).show_subnetpool(
            subnet_pool_id, **params)['subnetpool']
    except _client.NotFound as ex:
        raise NoSuchSubnetPool from ex


def create_subnet_pool(client: _client.NeutronClientType = None,
                       add_cleanup: bool = True,
                       **params) -> SubnetPoolType:
    subnet_pool = _client.neutron_client(client).create_subnetpool(
        body={'subnetpool': params})['subnetpool']
    if add_cleanup:
        tobiko.add_cleanup(
            cleanup_subnet_pool, subnet_pool=subnet_pool, client=client)
    return subnet_pool


def cleanup_subnet_pool(subnet_pool: SubnetPoolIdType,
                        client: _client.NeutronClientType = None):
    try:
        delete_subnet_pool(subnet_pool=subnet_pool, client=client)
    except NoSuchSubnetPool:
        pass


def delete_subnet_pool(subnet_pool: SubnetPoolIdType,
                       client: _client.NeutronClientType = None):
    subnet_pool_id = get_subnet_pool_id(subnet_pool)
    try:
        _client.neutron_client(client).delete_subnetpool(subnet_pool_id)
    except _client.NotFound as ex:
        raise NoSuchSubnetPool from ex


def list_subnet_pools(client: _client.NeutronClientType = None,
                      **params) -> tobiko.Selection[SubnetPoolType]:
    subnet_pools = _client.neutron_client(client).list_subnetpools(**params)
    if isinstance(subnet_pools, abc.Mapping):
        subnet_pools = subnet_pools['subnetpools']
    return tobiko.select(subnet_pools)


def find_subnet_pool(client: _client.NeutronClientType = None,
                     unique=False,
                     **params) -> SubnetPoolType:
    """Look for a subnet pool matching some values"""
    subnet_pools = list_subnet_pools(client=client, **params)
    if subnet_pools:
        if unique:
            return subnet_pools.unique
        else:
            return subnet_pools.first
    else:
        raise NoSuchSubnetPool


class NoSuchSubnetPool(tobiko.ObjectNotFound):
    message = "No such subnet pool found"



# Copyright 2020 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import functools
import io
import os
import re
import typing

from oslo_log import log
from packaging import version

import tobiko
from tobiko import config
from tobiko.shell import ansible
from tobiko.shell import sh
from tobiko.shell import ssh
from tobiko.tripleo import _overcloud
from tobiko.tripleo import _undercloud
from tobiko.tripleo import _config


LOG = log.getLogger(__name__)


def get_tripleo_ansible_inventory():
    inventory_file = get_tripleo_ansible_inventory_file()
    with io.open(inventory_file, 'rb') as fd:
        return tobiko.load_yaml(fd)


def has_tripleo_ansible_inventory() -> bool:
    inventory_file = get_tripleo_ansible_inventory_file()
    return inventory_file is not None


skip_if_missing_tripleo_ansible_inventory = \
    tobiko.skip_unless("Can't read TripleO Ansible inventory",
                       has_tripleo_ansible_inventory)


@functools.lru_cache()
def get_tripleo_ansible_inventory_file() -> typing.Optional[str]:
    if _undercloud.has_undercloud():
        inventory_file = config.CONF.tobiko.rhosp.inventory_file
        if inventory_file:
            inventory_file = tobiko.tobiko_config_path(inventory_file)
            fetch_tripleo_inventary_file(inventory_file=inventory_file)
            return inventory_file
    return None


def skip_unless_undercloud_has_ansible(min_version: float = None):
    reason = ("Ansible not found on undercloud host "
              f"(min_version={min_version})")
    return tobiko.skip_unless(reason=reason,
                              predicate=has_undercloud_ansible_version,
                              min_version=min_version)


def has_undercloud_ansible_version(min_version: float = None) -> bool:
    ssh_client = _undercloud.undercloud_ssh_client()
    try:
        output = sh.execute('ansible --version',
                            ssh_client=ssh_client).stdout
    except sh.ShellCommandFailed:
        LOG.debug("Error getting ansible version", exc_info=1)
        return False

    if min_version is not None:
        first_line = output.splitlines()[0]
        # obtain the ansible version with a regex
        # the expected output can have a format like these
        # ansible 2.9.27
        # ansible [core 2.12.2]
        search_version = re.search(r'(\d+\.)?(\d+\.)?(\*|\d+)', first_line)
        if not search_version:
            return False
        ansible_version_full = search_version.group(0)
        version_pair = ansible_version_full.split('.', 2)[:2]
        # compare versions rather than floats - example of a posible mistake:
        # version 2.12 is higher than version 2.9
        current_version = version.parse('.'.join(version_pair))
        if current_version < version.parse(str(min_version)):
            LOG.debug(f"Ansible version is < {min_version}:\n"
                      f"{first_line}\n")
            return False
    return True


READ_TRIPLEO_ANSIBLE_INVENTORY_SCRIPT = """
source {undercloud_rcfile} || exit 1

set -x

if [ "$OS_CLOUD" == "" ]; then
    # on some releases tripleo-ansible-inventory command is supported
    INVENTORY_FILE=$(mktemp tripleo-hosts-XXXXXXXXXX.yaml)
    tripleo-ansible-inventory --ansible_ssh_user "{overcloud_ssh_username}" \\
                              --static-yaml-inventory "$INVENTORY_FILE"
    RC=$?

    if [ $RC == 0 ]; then
        cat "$INVENTORY_FILE"
    fi
    rm -fR "$INVENTORY_FILE"
    exit $RC
fi

# on latest releases, the updated inventory can always be found into the
# following path
TAI_DIR=~/overcloud-deploy/overcloud/config-download/overcloud
cat $TAI_DIR/tripleo-ansible-inventory.yaml
"""


def read_tripleo_ansible_inventory():
    tripleo = _config.get_tripleo_config()
    ssh_client = _undercloud.undercloud_ssh_client()
    script = READ_TRIPLEO_ANSIBLE_INVENTORY_SCRIPT.format(
        undercloud_rcfile=tripleo.undercloud_rcfile[0],
        overcloud_ssh_username=_overcloud.get_overcloud_ssh_username())
    return sh.execute('/bin/bash', stdin=script, ssh_client=ssh_client).stdout


def fetch_tripleo_inventary_file(inventory_file: str):
    content = read_tripleo_ansible_inventory()
    tobiko.makedirs(os.path.dirname(inventory_file))
    with io.open(inventory_file, 'w') as fd:
        fd.write(content)


class UndercloudAnsiblePlaybook(ansible.AnsiblePlaybook):

    @property
    def ssh_client(self) -> ssh.SSHClientType:
        return _undercloud.undercloud_ssh_client()

    def _ensure_inventory_files(self, *inventory_filenames: str) \
            -> typing.List[str]:
        inventory_file = get_tripleo_ansible_inventory_file()
        if inventory_file is not None:
            inventory_filenames += (inventory_file,)
        return super()._ensure_inventory_files(*inventory_filenames)


def undercloud_ansible_playbook() -> UndercloudAnsiblePlaybook:
    return tobiko.get_fixture(UndercloudAnsiblePlaybook)


def run_playbook_from_undercloud(
        command: sh.ShellCommand = None,
        playbook: str = None,
        playbook_dirname: str = None,
        playbook_filename: str = None,
        inventory_filenames: typing.Iterable[str] = None,
        playbook_files: typing.Iterable[str] = None,
        requirements_files: typing.Iterable[str] = None,
        roles: typing.Iterable[str] = None,
        roles_path: typing.Iterable[str] = None,
        verbosity: int = None):
    return undercloud_ansible_playbook().run_playbook(
        command=command,
        playbook=playbook,
        playbook_dirname=playbook_dirname,
        playbook_filename=playbook_filename,
        inventory_filenames=inventory_filenames,
        playbook_files=playbook_files,
        requirements_files=requirements_files,
        roles=roles,
        roles_path=roles_path,
        verbosity=verbosity)


def setup_undercloud_ansible_playbook():
    if _undercloud.has_undercloud():
        ansible.register_ansible_playbook(
            undercloud_ansible_playbook())
